##############################################################################
#
# Copyright (c) 2009 Projekt01 GmbH and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""A pymongo client with environment based setup
$Id: client.py 5178 2025-03-18 15:08:21Z felipe.souza $
"""
from __future__ import absolute_import
from builtins import str
__docformat__ = "reStructuredText"

import pymongo


###############################################################################
#
# pymongo client setup

client = None

# TODO: implement all options defined in pymongo/common.py
# TODO: support **kwargs and update options at the end for override existing
#       or define undefined options given from new mongo versions

def getMongoClient(settings=None, connect=True, retry=True):
    global client
    if client is not None:
        # force to close monitor thread
        client.close()
    # setup client
    if settings is None:
        from m01.mongo import settings
    uri = settings.MONGODB_URI
    if uri == 'testing' or settings.P01_CDN_RECIPE_PROCESS or \
        settings.I18N_EXTRACT:
        # testing or resource extraction, just a dummy client using testing port
        return pymongo.MongoClient('localhost:45017')
    else:
        uri = str(uri)
        if settings.MONGODB_SSL and \
            settings.MONGODB_SSL_CERT_REQUIRED is None:
            # if undefined use ssl.CERT_REQUIRED
            import ssl
            ssl_cert_reqs = ssl.CERT_REQUIRED
        else:
            ssl_cert_reqs = settings.MONGODB_SSL_CERT_REQUIRED
        if not settings.MONGODB_CONNECT:
            connect = False
        options = {
            'connect': connect,
            'tz_aware': settings.MONGODB_TZ_AWARE,
        }
        if settings.MONGODB_REPLICA_SET:
            options['replicaSet'] = settings.MONGODB_REPLICA_SET
        if settings.MONGODB_APP_NAME:
            options['appname'] = settings.MONGODB_APP_NAME
        if settings.MONGODB_MIN_POOL_SIZE:
            options['minPoolSize'] = settings.MONGODB_MIN_POOL_SIZE
        if settings.MONGODB_MAX_POOL_SIZE:
            options['maxPoolSize'] = settings.MONGODB_MAX_POOL_SIZE
        if settings.MONGODB_CONNECT_TIMEOUT:
            options['connectTimeoutMS'] = \
                settings.MONGODB_CONNECT_TIMEOUT
        if settings.MONGODB_SERVER_SELECTION_TIMEOUT:
            options['serverSelectionTimeoutMS'] = \
                settings.MONGODB_SERVER_SELECTION_TIMEOUT
        if settings.MONGODB_SOCKET_TIMEOUT:
            options['socketTimeoutMS'] = settings.MONGODB_SOCKET_TIMEOUT
        if settings.MONGODB_WAIT_QUEUE_TIMEOUT:
            options['waitQueueTimeoutMS'] = \
                settings.MONGODB_WAIT_QUEUE_TIMEOUT
        if settings.MONGODB_MAX_IDLE_TIME_MS:
            options['maxIdleTimeMS'] = \
                settings.MONGODB_MAX_IDLE_TIME_MS
        if settings.MONGODB_WAIT_QUEUE_MULTIPLE:
            options['waitQueueMultiple'] = \
                settings.MONGODB_WAIT_QUEUE_MULTIPLE
        if settings.MONGODB_HEARTBEAT_FREQUENCY:
            options['heartbeatFrequencyMS'] = \
                settings.MONGODB_HEARTBEAT_FREQUENCY
        # auth options
        if settings.MONGODB_AUTH_MECHNISM:
            options['authMechanism'] = settings.MONGODB_AUTH_MECHNISM
        if settings.MONGODB_USERNAME:
            options['username'] = settings.MONGODB_USERNAME
        if settings.MONGODB_PASSWORD:
            options['password'] = settings.MONGODB_PASSWORD
        # ssl options
        if settings.MONGODB_SSL:
            options['ssl'] = settings.MONGODB_SSL
            if ssl_cert_reqs:
                options['ssl_cert_reqs'] = ssl_cert_reqs
            if settings.MONGODB_CA_CERT:
                options['ssl_ca_certs'] = settings.MONGODB_CA_CERT
            if settings.MONGODB_CLIENT_CERT:
                options['ssl_certfile'] = settings.MONGODB_CLIENT_CERT
            if settings.MONGODB_CLIENT_CERT_KEY_FILE:
                options['ssl_keyfile'] = settings.MONGODB_CLIENT_CERT_KEY_FILE
            if settings.MONGODB_PEM_PASSPHRASE:
                options['ssl_pem_passphrase'] = settings.MONGODB_PEM_PASSPHRASE
            if settings.MONGODB_REVOCATION_LIST:
                options['ssl_crlfile'] = settings.MONGODB_REVOCATION_LIST
            if settings.MONGODB_SSL_MATCH_HOSTNAME:
                options['ssl_match_hostname'] = \
                    settings.MONGODB_SSL_MATCH_HOSTNAME
        # tls options
        if settings.MONGODB_TLS:
            options['tls'] = settings.MONGODB_TLS
            if settings.MONGODB_TLS_CA_FILE:
                options['tlsCAFile'] = settings.MONGODB_TLS_CA_FILE
            if settings.MONGODB_TLS_CERT_KEY_FILE:
                options['tlsCertificateKeyFile'] = \
                    settings.MONGODB_TLS_CERT_KEY_FILE
            if settings.MONGODB_TLS_CERT_KEY_FILE_PASSWORD:
                options['tlsCertificateKeyFilePassword'] = \
                    settings.MONGODB_TLS_CERT_KEY_FILE_PASSWORD
            if settings.MONGODB_TLS_CRL_FILE:
                options['tlsCRLFile'] = settings.MONGODB_TLS_CRL_FILE
            if settings.MONGODB_TLS_ALLOW_INVALID_CERT:
                options['tlsAllowInvalidCertificates'] = \
                    settings.MONGODB_TLS_ALLOW_INVALID_CERT
            if settings.MONGODB_TLS_ALLOW_INVALID_HOSTNAME:
                options['tlsAllowInvalidHostnames'] = \
                    settings.MONGODB_TLS_ALLOW_INVALID_HOSTNAME
            if settings.MONGODB_TLS_CERTIFICATE_SELECTOR:
                options['tlsCertificateSelector'] = \
                    settings.MONGODB_TLS_CERTIFICATE_SELECTOR
        try:
            client = pymongo.MongoClient(uri, **options)
            if connect:
                client.admin.command('ismaster')
            return client
        except pymongo.errors.ConnectionFailure as e:
            # ensure client, try again without connect
            if retry:
                return getMongoClient(settings, connect=False, retry=False)
            else:
                raise e



###############################################################################
#
# Copyright (c) 2011 Projekt01 GmbH and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
###############################################################################
"""
$Id:$
"""
from __future__ import absolute_import
__docformat__ = "reStructuredText"

import calendar
import time
import datetime
import struct
import threading
import zope.component

import bson.son
import bson.objectid

from m01.mongo.tz_util import UTC


###############################################################################
#
# threading

LOCAL = threading.local()

def clearThreadLocalCache(event=None):
    """A subscriber to EndRequestEvent

    Cleans up the thread local cache on each end request.
    """
    for key in list(LOCAL.__dict__.keys()):
        del LOCAL.__dict__[key]


###############################################################################
#
# datetime hook

_tzmarker = object()

# testing hook for datetime
# allows to set the current date/time
NOW = None

def now(tz=_tzmarker):
    global NOW
    if tz is _tzmarker:
        tz = UTC
    if NOW is not None:
        return NOW(tz)
    else:
        return datetime.datetime.now(tz)

def setNOW(value=None):
    global NOW
    if value is None:
        NOW = None
    elif callable(value):
        NOW = value
    else:
        NOW = lambda tz: value.replace(tzinfo=tz)

def getNOW():
    global NOW
    return NOW

def today():
    return now().date()


###############################################################################
#
# SON to dict converter
def dictify(data):
    """Recursive replace SON items with dict in the given data structure.

    Compared to the SON.to_dict method, this method will also handle tuples
    and keep them intact.

    """
    if isinstance(data, bson.son.SON):
        data = dict(data)
    if isinstance(data, dict):
        d = {}
        for k, v in list(data.items()):
            # replace nested SON items
            d[k] = dictify(v)
    elif isinstance(data, (tuple, list)):
        d = []
        for v in data:
            # replace nested SON items
            d.append(dictify(v))
        if isinstance(data, tuple):
            # keep tuples intact
            d = tuple(d)
    else:
        d = data
    return d


###############################################################################
#
# object id generation

def getObjectId(secs=0):
    """Knows how to generate similar ObjectId based on integer (counter)

    Note: this method can get used if you need to define similar ObjectId
    in a non persistent environment if need to bootstrap mongo containers.
    """
    time_tuple = time.gmtime(secs)
    ts = calendar.timegm(time_tuple)
    oid = struct.pack(">i", int(ts)) + b"\x00" * 8
    return bson.objectid.ObjectId(oid)


def getObjectIdByTimeStr(tStr, frmt="%Y-%m-%d %H:%M:%S"):
    """Knows how to generate similar ObjectId based on a time string

    The time string format used by default is ``%Y-%m-%d %H:%M:%S``.
    Use the current development time which could prevent duplicated
    ObjectId. At least some kind of ;-)
    """
    time.strptime(tStr, frmt)
    ts = time.mktime(tStr)
    oid = struct.pack(">i", int(ts)) + b"\x00" * 8
    return bson.objectid.ObjectId(oid)



###############################################################################
#
# Copyright (c) 2011 Projekt01 GmbH and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
###############################################################################
"""
$Id:$
"""
from __future__ import absolute_import
from __future__ import print_function

__docformat__ = "reStructuredText"

import unittest
import doctest

import m01.mongo.testing


def test_suite():
    suites = []
    append = suites.append

    # fake database tests. our fake database does not support ensure_index and
    # special prefixed commands like ($near). This menas we can't run all tests
    # with our fake database
    fakeTestNames = [
        'client.txt',
        'container.txt',
        'object.txt',
        'shared.txt',
        'storage.txt',
        'usecase.txt',
        'util.txt',
        'zope.schema.txt',
    ]
    for name in fakeTestNames:
        append(
            doctest.DocFileSuite(name,
                globs={'print_function': print_function,
                    'absolute_import': absolute_import},
                setUp=m01.mongo.testing.setUpFakeMongo,
                tearDown=m01.mongo.testing.tearDownFakeMongo,
                optionflags=doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS,
                checker=m01.mongo.testing.reNormalizer),
        )
    append(
        doctest.DocFileSuite('testing.txt',
            globs={'print_function': print_function,
                'absolute_import': absolute_import},
            optionflags=doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS,
            checker=m01.mongo.testing.reNormalizer),
    )

    # real mongo database tests using m01.stub using level 2 tests (--all)
    testNames = [
        '../README.txt',
        'batching.txt',
        'container.txt',
        'geo.txt',
        'geopoint.txt',
        'object.txt',
        'shared.txt',
        'storage.txt',
        'usecase.txt',
        'zope.schema.txt',
    ]
    for name in testNames:
        suite = unittest.TestSuite((
            doctest.DocFileSuite(name,
                globs={'print_function': print_function,
                    'absolute_import': absolute_import},
                setUp=m01.mongo.testing.setUpStubMongo,
                tearDown=m01.mongo.testing.tearDownStubMongo,
                optionflags=doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS,
                checker=m01.mongo.testing.reNormalizer),
        ))
        suite.level = 2
        append(suite)

    # return test suite
    return unittest.TestSuite(suites)


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



###############################################################################
#
# Copyright (c) 2012 Projekt01 GmbH and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
###############################################################################
"""
$Id: tz_util.py 5178 2025-03-18 15:08:21Z felipe.souza $
"""
from __future__ import absolute_import
__docformat__ = "reStructuredText"

import datetime

# UTC support
ZERO = datetime.timedelta(0)


class FixedOffsetBase(datetime.tzinfo):
    """Fixed offset base class."""

    def __init__(self, offset, name):
        if isinstance(offset, int):
            offset = datetime.timedelta(minutes=offset)
        self.__offset = offset
        self.__name = name

    def utcoffset(self, dt):
        return self.__offset

    def tzname(self, dt):
        return self.__name

    def dst(self, dt):
        return ZERO

    def __repr__(self):
        return self.__name


class FixedOffset(FixedOffsetBase):
    """Pickable implementation of fixed offset used for deepcopy."""

    def __init__(self, offset=None, name=None):
        FixedOffsetBase.__init__(self, offset, name)

# Fixed offset timezone representing UTC
UTC = FixedOffset(0, "UTC")



###############################################################################
#
# Copyright (c) 2011 Projekt01 GmbH and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
###############################################################################
"""
$Id:$
"""
from __future__ import absolute_import
from six import string_types
import six
__docformat__ = "reStructuredText"

import re
import datetime

import bson.binary

import zope.interface
import zope.schema._field
import zope.schema.interfaces

from m01.mongo import interfaces
from m01.mongo.tz_util import UTC
import m01.mongo.util


@zope.interface.implementer(interfaces.IMongoList)
class MongoList(zope.schema._field.AbstractCollection):
    """A field representing a List, internaly using a traversable container."""

    _type = (m01.mongo.util.MongoItemsData, m01.mongo.util.MongoListData, list)

    def __init__(self, value_type=None, unique=False, **kw):
        # whine if value_type is not defined because we need it for identify
        # the list data types e.g. MongoItems or MongoListData
        if value_type is None:
            raise ValueError("Missing 'value_type' in field")
        super(MongoList, self).__init__(value_type, unique, **kw)

    def set(self, obj, value):
        super(MongoList, self).set(obj, value)
        try:
            # locate the given item, but note, this set method is bypassed if we
            # use MongoFieldProperty which stores the value in __dict__
            value.__parent__ = obj
            value.__name__ = self.__name__
        except AttributeError as e:
            pass


@zope.interface.implementer(interfaces.IMongoBinary)
class MongoBinary(zope.schema._field.Bytes):
    """Field which stores bson.binary.Binary data"""

    _type = bson.binary.Binary

    def set(self, obj, value):
        if isinstance(value, six.string_types):
            value = bson.binary.Binary(value)
        super(MongoBinary, self).set(obj, value)


@zope.interface.implementer(interfaces.IMongoDate)
class MongoDate(zope.schema._field.Date):
    """A field representing a date, internaly using an ordinal integer value."""

    _type = (int, datetime.date)

    def set(self, obj, value):
        if isinstance(value, int):
            value = datetime.date.fromordinal(value)
        super(MongoDate, self).set(obj, value)


@zope.interface.implementer(interfaces.IMongoDatetime)
class MongoDatetime(zope.schema._field.Datetime):
    """A field representing a timzone aware datetime."""

    _type = datetime.datetime
    tzinfo = UTC

    def __init__(self, *args, **kw):
        self.tzinfo = kw.pop("tzinfo", UTC)
        super(MongoDatetime, self).__init__(**kw)

    def _validate(self, value):
        super(MongoDatetime, self)._validate(value)
        if not isinstance(value, datetime.datetime):
            raise zope.schema.interfaces.WrongType(value, self._type,
                self.__name__)

    def set(self, context, value):
        if isinstance(value, datetime.datetime):
            value = value.replace(tzinfo=self.tzinfo)
        super(MongoDatetime, self).set(context, value)


# zope.schema string based fields using unicode
_isuri = re.compile(
    # scheme
    r"[a-zA-z0-9+.-]+:"
    # non space (should be pickier)
    r"\S*$").match

@zope.interface.implementer(zope.schema.interfaces.IURI)
class URI(zope.schema.TextLine):
    """URI schema field using unicode as stored value"""

    _type = str

    def _validate(self, value):
        super(URI, self)._validate(value)
        if _isuri(value):
            return
        raise zope.schema.interfaces.InvalidURI(value)



###############################################################################
#
# Copyright (c) 2011 Projekt01 GmbH and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
###############################################################################
"""
$Id:$
"""
from __future__ import absolute_import
__docformat__ = "reStructuredText"

import zope.interface
import zope.i18nmessageid
import zope.schema
import zope.location.interfaces
import zope.container.interfaces
import zope.schema.interfaces

from m01.mongo.tz_util import UTC

_ = zope.i18nmessageid.MessageFactory('p01')


class IMongoList(zope.schema.interfaces.IList):
    """Field containing a value that implements the API of a conventional
    Python list used in mogo item list.
    """


# class IMongoSubObjectConverter(zope.schema.interfaces.IObject):
#     """Mongo object converter field"""


class IMongoBinary(zope.schema.interfaces.IBytes):
    """Field which stores bson.binary.Binary data"""


class IMongoDate(zope.schema.interfaces.IDate):
    """Field which stores an ordinal integer as date."""


class IMongoDatetime(zope.schema.interfaces.IDatetime):
    """A field representing a timzone aware datetime."""

    tzinfo = zope.schema.Field(
        title=_("Timezone info"),
        description=_("Timezone info"),
        default=UTC
        )


class IMongoListData(zope.location.interfaces.ILocation):
    """Simple item list supporting ILocation"""

    def clear():
        """Clear data"""

    def dump():
        """Dump the list to json data"""

    def append(item):
        """Append an item to the list."""

    def insert(i, item):
        """Insert an item at the given position."""

    def pop(i=-1):
        """Pop an item from the list."""

    def remove(item):
        """Remove an item from the list."""

    def sort(*args, **kwargs):
        """Sort the list"""

    def __len__():
        """Returns the lenght."""

    def __contains__(item):
        """Returns True if the list contains the given item otherwise False."""

    def __getitem__(i):
        """Get an item from the list."""

    def __setitem__(i, item):
        """Add an item to the list."""

    def __delitem__(i):
        """Remove an item from the list."""

    def __cmp__(other):
        """compare"""

    def __repr__():
        """representation"""


class IMongoItemsData(zope.location.interfaces.ILocation):
    """Mongo item list supporting ILocation"""

    def clear():
        """Clear data"""

    def dump():
        """Dump the list to json data"""

    def locate(value):
        """Locate an object with parent but only if locatable."""

    def locateAndReturn(obj):
        """Locate and return an object with parent but only if locatable."""

    def reLocate(__parent__):
        """Update location, this is needed if we use LocationProxy."""

    def order(orderNames):
        """Order items by the given name order"""

    def fetch(idx):
        """Returns the item at the given position. Same as <type list>[idx]"""

    def append(item):
        """Append new item"""

    def insert(i, item):
        """Insert new item at given order"""

    def remove(item):
        """Remove an item"""

    def __len__():
        """Returns the lenght."""

    def __contains__(key):
        """Return True if item is a part of data"""

    def __getitem__(key):
        """Get item"""

    def __setitem__(key, item):
        """Set item"""

    def __delitem__(key):
        """Remove an item"""

    def get(key, default=None):
        """Get item by key"""

    def keys():
        """Get keys"""

    def values():
        """Get values"""

    def items():
        """Get items"""

    def __iter__():
        """Get iterator"""

    def __cmp__(other):
        """compare"""

    def __repr__():
        """representation"""


class ISecurityAware(zope.interface.Interface):
    """Security aware mixin."""

    # principal permission manager
    _ppmrow = zope.schema.Dict(
        title=u'Principal Permission byrow',
        default={})
    _ppmcol = zope.schema.Dict(
        title=u'Principal Permission bycol',
        default={})

    # principal role manager
    _prmrow = zope.schema.Dict(
        title=u'Principal Role byrow',
        default={})
    _prmcol = zope.schema.Dict(
        title=u'Principal Role bycol',
        default={})

    # role permission manager
    _rpmrow = zope.schema.Dict(
        title=u'Role Permission byrow',
        default={})
    _rpmcol = zope.schema.Dict(
        title=u'Role Permission bycol',
        default={})

#    principalPermissionManager = zope.interface.Attribute(
#        """Principal Permission Manager""")
#    principalRoleManager = zope.interface.Attribute(
#        """Principal Role Manager""")
#    rolePermissionManager = zope.interface.Attribute(
#        """Role Permission Manager""")


class IMongoTransactionAware(zope.interface.Interface):
    """Mongo transaction support."""

    # transaction data manager support methods
    def begin():
        """Beginn commit a transaction handling."""

    def commit():
        """Start commit transaction handling."""

    def vote():
        """Process MongoDB insert and update calls.

        Note, if something fails, this could leave inconsistent data in MongoDB.
        """

    def abort():
        """Abort transaction handling."""

    def finish():
        """Finish transaction commit."""


class IMongoCollectionAware(zope.interface.Interface):
    """Mongo item loader."""

    collection = zope.interface.Attribute("""A mongoDB collection.""")

    def doLoad(data):
        """Prepare data and load them into IMongoItem."""

    def load(data):
        """Load data into an IMongoStorageItem,

        Override this method if you need another mongo item.
        """


class IMongoBatchAware(zope.interface.Interface):
    """Batching support"""

    def getBatchData(query=None, page=1, size=25, sortName=None,
        sortOrder=None, searchText=None, fields=None, skipFilter=False):
        """Returns batched mongo data, current page, total items and page size.
        """

# setup, convert, dump
class ISetup(zope.interface.Interface):
    """Setup mixin"""

    def setup(data, dumpOriginalData=False):
        """Update object with given data key/values.

        The method returns the original data as dict including the original
        _m_changed and _version values. This could be usefull for revert
        changes.

        Note: this method can be used for update an object and commit changes
        ASAP to MongoDB. See update method in IMongoHandler for more info

        Note: this will change the _m_changed state if some key/value are using
        a MongoFieldProperty!
        """

class IConvert(zope.interface.Interface):
    """Setup mixin"""

    def convert(key, value):
        """This convert method knows how to handle nested converters.

        Normaly a converter can convert attribute values. If the attribute
        value is a list of items, then you can use amethod whihc knows how to
        handle each value type. Or you can use this converter method and
        define nested converter. If so, juset define a dict with _type/converter
        for a converter attr name.
        """

class IDump(zope.interface.Interface):
    """Setup mixin"""

    def dump(dumpNames=None, dictify=False):
        """Dump the object to bson.son.SON instance.

        Optional the values filtered by the given dumpNames can get dumped to
        a dict. This is usefull if you need partial data from an object as
        key/value dict.

        NOTE: the dictify argument forces to recoursive convert bson.son.SON
        instances to a plain dict.

        You can also use the dictify method from m01.mongo if you like to
        replace the SON instances with plain dicts. This is sometimes required
        e.g. if you need to convert the data to a json string.

        """


class ICreatedModified(zope.interface.Interface):
    """Created and modified mixin"""

    created = zope.schema.Datetime(
        title=_(u'created'),
        description=_(u'created'),
        default=None,
        required=True)

    modified = zope.schema.Datetime(
        title=_(u'modified'),
        description=_(u'modified'),
        default=None,
        required=False)


# mongo item
class IMongoSubItem(ISetup, IConvert, IDump, ICreatedModified):
    """Mongo sub item

    NOTE: this interface is the core interface for mongo sub items but this
    interface doens't provide ILocation or IContained. In most use case you
    need to provide ILocation in your own inherited interface. The missing
    ILocation interface in this interface makes it simpler for define correct
    permissions in ZCML.
    """

    _m_changed = zope.interface.Attribute(
        """The persistent state of the object.

        This is one of:

        None -- The object is loaded from the MongoDB

        False -- The obj get updated within the update method and set the False

        True -- The object has been modified since it was last saved

        """)


class IMongoUpdateMixin(zope.interface.Interface):
    """Optional update method mixin class usfull for adhoc key/value updates.
    """

    def update(data, raiseErrors=False):
        """Update an object with given data without using the built in
        transaction handling.

        This method is usefull for update key/values given in a data dict
        which need to get updated during processing. On a successfull update,
        True is returnd otherwise False.

        If raiseErrors is set to True, an error get raised.

        We will revert the changes on errors, doesn't matter if an error happens
        and get skipped. Our transaction is trying to cleanup/store the
        original data during commit.

        We also update the _version in MongoDB and the object itself and set
        the _m_changed marker to False which is a marker for calling this
        update method.

        Note, changed key/values before calling this method get also stored
        in MongoDB. And we do only revert to the data to the current state
        before we called the update method. This means you have to take care
        or you can messup your data.

        It is important to know that if you do not raise Errors and the return
        value is False, the data did not get updated.
        """


class IMongoParent(zope.interface.Interface):
    """Offers mongo object id (_mpid) support

    The _mpid get stored as _pid in a mongo item and is used as a __parent__
    reference.

    Normaly a mongo object _id get used as _mpid. By default the _id and
    _mpid are None. If a IMongoParent is used as a mixn in an IMongoItem, the
    mongo item will set this _mpid as the items _pid.
    """

    _id = zope.schema.Field(u"Mongo ObjectId.")
    _mpid = zope.schema.Field(u"Referencable mongo __parent__ ObjectId")


class IMongoParentAware(zope.interface.Interface):
    """Offers __parent__ id (_pid) reference support

    Normaly the IMongoParent _mpid get used as _pid.
    """

    _pid = zope.schema.Field(u"Items __parent__ ObjectId.")


class IMongoItem(ISetup, IConvert, IDump, ICreatedModified):
    """Mongo storage item schema without ILocation.

    NOTE: this interface is the core interface for mongo sub items but this
    interface doens't provide ILocation or IContained. In most use case you
    need to provide ILocation in your own inherited interface. The missing
    ILocation interface in this interface makes it simpler for define correct
    permissions in ZCML.
    """

    _id = zope.schema.Field(u"Mongo ObjectId.")

    _m_changed = zope.interface.Attribute(
        """The persistent state of the object.

        This is one of:

        None -- The object is loaded from the MongoDB

        False -- The obj get updated within the update method and set the False

        True -- The object has been modified since it was last saved

        """)

    _m_independent = zope.interface.Attribute("""Independent state""")

    _type = zope.schema.TextLine(
        title=u'Type',
        description=u'Mongo item type',
        default=None,
        required=True)

    _version = zope.schema.Int(
        title=u'Version',
        description=u'Mongo item version',
        default=0,
        required=True)

    def notifyRemove():
        """Notifies an item before dumped and removed from MongoDB.

        This allows us to cleanup the object before we remove from MongoDB.
        e.g. remove references from other databases. This let us add a remove
        marker for track file system references in m01.mongofs.
        """

class IMongoSubObject(ISetup, IConvert, IDump):
    """Mongo item base class.

    Note: only MongoFieldProperty attributes get observed and updated by our
    transaction handling. If you change other attributes, you have to set
    _m_changed to True by yourself.

    Note: We do not support manipulators known from pymongo. We use a converter
    pattern for ding the convertion, see converters
    """

    _m_changed = zope.interface.Attribute(
        """The persistent state of the object.

        This is one of:

        None -- The object is loaded from the MongoDB

        False -- The obj get updated within the update method and set the False

        True -- The object has been modified since it was last saved

        """)

    _type = zope.schema.TextLine(
        title=u'Type',
        description=u'Mongo item type',
        default=None,
        required=True)


class IMongoStorageItem(IMongoItem):
    """Mongo storage item."""


class IMongoContainerItem(IMongoItem):
    """Mongo container item."""


class ISecureMongoStorageItem(IMongoStorageItem, ISecurityAware):
    """Secure mongo storage item."""


class ISecureMongoContainerItem(IMongoContainerItem, ISecurityAware):
    """Secure mongo container item."""


class IMongoObjectAware(zope.interface.Interface):
    """Object that contains IMongoObject properties.

    This interface is optional. It just shows what an object must implement
    which uses a MongoObjectProperty for one or more attributes. The defined
    _moid is, together with a double point followed by the attribute name, used
    as a parent reference id in a MongoObject and exposed as _oid.

    An _moid could be a MongoDB _id or something like:

    ser43130chke34

    if such an object defines an attribute called myattr, the MongoObject _oid
    looks like:

    ser43130chke34:myattr

    The MongoFieldProperty is responsible for setup such _oid values based on
    the given instance _moid and field name.

    Attention: the _oid must be unique in your application, or at least unique
    in the MongoDB which the MongoObject get stored. Otherwise an
    IMongoAwareObject could get a wrong attribute value. Except if you like to
    share a MongoObject within different MongoObject __parent__ by using a
    shared _moid.

    We explicitly choosed this naming reference concept because this allows us
    to add new MongoObject without to write into the __parent__ object. Another
    concept would be to write the MongoObject _id into the MongoAwareObject
    which would force a write access to both the MongoAwareObject and
    MongoObject.

    """

    _moid = zope.schema.TextLine(
        title=u'Unique object ID',
        description=u'Unique ID used as MongoObject __parent__ reference',
        default=None,
        required=True,
        readonly=True)


class IMongoObject(IMongoItem, IMongoTransactionAware):
    """Mongo object without ILocation

    This mongo (item) object is observed by it's own transaction data manager.

    NOTE: this interface is the core interface for mongo sub items but this
    interface doens't provide ILocation or IContained. In most use case you
    need to provide ILocation in your own inherited interface. The missing
    ILocation interface in this interface makes it simpler for define correct
    permissions in ZCML.
    """

    _oid = zope.schema.TextLine(
        title=u'Mongo parent object ID including attribute name',
        description=u'Mongo parent object ID with including attribute name',
        default=None,
        required=True,
        readonly=True)

    # this is new since 0.6.0, the MongoObjectProperty will converte the items
    # during load and save them by it's _oid which does not change
    _field = zope.schema.TextLine(
        title=u'Name of parent field where this item is stored in',
        description=u'Name of parent field where this item is stored in',
        default=None,
        required=True)

    removed = zope.schema.Bool(
        title=u'Removed marker',
        description=u'Removed marker',
        default=False,
        required=True)

    def getCollection(parent):
        """Returns the right collection where the mongo object get stored.

        Note, this method must be a classmethod whihc makes it work with
        MongoObjectProperty
        """


# mongo mapping
class IMongoStorage(IMongoCollectionAware, IMongoBatchAware,
    zope.container.interfaces.IReadContainer):
    """Mongo storage without ILocation"""

    def add(obj):
        """items"""

    def __delitem__(name):
        """Delete the named object from the container."""


class IMongoContainer(IMongoCollectionAware, IMongoBatchAware,
    zope.container.interfaces.IContainer):
    """Mongo container without ILocation"""


# geo location support
class IGeoSchema(IDump):
    """Geo object base schema"""

    lon = zope.schema.Float(
        title=_(u'longitude'),
        description=_(u'longitude'),
        required=True)

    lat = zope.schema.Float(
        title=_(u'latitude'),
        description=_(u'latitude'),
        required=True)


class IGeoLocation(IGeoSchema):
    """Geo location data wrapper, providing preserved lon/lat value order
    and supporting _m_changed marker concept.

    This item must get defined with a zope.schema.Object field and the
    MongoGeoLocationProperty property. for more info see geo.txt and testing.py

    IMPORTANT:
    we use the order latitude, longitude and (NOT lon, lat) like described in
    mongodb documentation

    A more detailed description about the lon/lat order thema:

    looking at the Geographic Coordinate System, you'll notice that longitude
    counts left to right, while latitude counts bottom to top. Translating that
    into x/y this would mean the Longitude/Latitude approach is the only one
    that makes sense logically.

    In english, coordinates are usually referred to as latitude/longitude in a
    term. This is due to 2 reasons: a precedents o in the alphabet and in
    Navigation, in the days of sextants Latitude was easier, and more exactly
    determined, so that was the first thing to determine/write down.

    From OGC Specs, lon/lat is the correct way to interpret them. You can
    easily see that from how those that where participants in creating the
    Specs (Oracle/Informix/ESRI/IBM) implemented them, and those that did
    their own stuff (MS/Google).

    Note: The GeoLocation is registered with zope.Public permission. This is
    fine for ost use case since you will protect the property which is storing
    such a GeoLocation item. There is no need to explicit protect the lon/lat
    attributes. Just make sure npbody with the correct permission can access
    the GeoLocation item itself.

    """


class IGeoPoint(IGeoSchema):
    """Geo Point data wrapper, providing preserved lon/lat value order
    and supporting _m_changed marker concept.

    This item must get defined with a zope.schema.Object field and the
    MongoGeoPointProperty property. For more info see geo.txt and testing.py

    This implementation uses the order longitude, latitude as described in
    mongodb documentation because of unnamed lng/lat list values.

    """

    type = zope.schema.TextLine(
        title=u'Geo object type',
        description=u'Geo object type',
        default=u'Point',
        readonly=True,
        required=True)



###############################################################################
#
# Copyright (c) 2011 Projekt01 GmbH and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
###############################################################################
"""
$Id:$
"""
from __future__ import absolute_import
from builtins import object
from zope.interface import implementer
__docformat__ = "reStructuredText"

import zope.interface
import zope.component
from zope.security.proxy import removeSecurityProxy
from zope.securitypolicy.interfaces import IPrincipalPermissionManager
from zope.securitypolicy.interfaces import IPrincipalRoleManager
from zope.securitypolicy.interfaces import IRolePermissionManager
from zope.securitypolicy.interfaces import IGrantInfo
from zope.securitypolicy.interfaces import Allow
from zope.securitypolicy.interfaces import Deny
from zope.securitypolicy.interfaces import Unset
from zope.security.management import queryInteraction

from m01.mongo import interfaces


class SecurityMap(object):
    """SecurityMap adapter base class."""

    def __init__(self, context):
        self.context = context

    def __bool__(self):
        return bool(self._byrow)

    def addCell(self, rowentry, colentry, value):
        row = self._byrow.get(rowentry)
        if row:
            if row.get(colentry) is value:
                return False
        else:
            row = self._byrow[rowentry] = {}

        col = self._bycol.get(colentry)
        if not col:
            col = self._bycol[colentry] = {}

        row[colentry] = value
        col[rowentry] = value

        self._invalidated_interaction_cache()
        self.context._m_changed = True
        return True

    def _invalidated_interaction_cache(self):
        # Invalidate this threads interaction cache
        interaction = queryInteraction()
        if interaction is not None:
            try:
                invalidate_cache = interaction.invalidate_cache
            except AttributeError:
                pass
            else:
                invalidate_cache()

    def delCell(self, rowentry, colentry):
        row = self._byrow.get(rowentry)
        if row and (colentry in row):
            del row[colentry]
            if not row:
                del self._byrow[rowentry]
            col = self._bycol[colentry]
            del col[rowentry]
            if not col:
                del self._bycol[colentry]

            self._invalidated_interaction_cache()
            self.context._m_changed = True
            return True

        return False

    def queryCell(self, rowentry, colentry, default=None):
        row = self._byrow.get(rowentry)
        if row:
            return row.get(colentry, default)
        else:
            return default

    def getCell(self, rowentry, colentry):
        marker = object()
        cell = self.queryCell(rowentry, colentry, marker)
        if cell is marker:
            raise KeyError('Not a valid row and column pair.')
        return cell

    def getRow(self, rowentry):
        row = self._byrow.get(rowentry)
        if row:
            return list(row.items())
        else:
            return []

    def getCol(self, colentry):
        col = self._bycol.get(colentry)
        if col:
            return list(col.items())
        else:
            return []

    def getAllCells(self):
        res = []
        for r in list(self._byrow.keys()):
            for c in list(self._byrow[r].items()):
                res.append((r,) + c)
        return res


class PPMSecurityMap(SecurityMap):

    @property
    def _byrow(self):
        return self.context._ppmrow

    @property
    def _bycol(self):
        return self.context._ppmcol


class PRMSecurityMap(SecurityMap):

    @property
    def _byrow(self):
        return self.context._prmrow

    @property
    def _bycol(self):
        return self.context._prmcol


class RPMSecurityMap(SecurityMap):

    @property
    def _byrow(self):
        return self.context._rpmrow

    @property
    def _bycol(self):
        return self.context._rpmcol


@implementer(IPrincipalPermissionManager)
class PrincipalPermissionManager(PPMSecurityMap):
    """Mappings between principals and permissions."""


    def grantPermissionToPrincipal(self, permission_id, principal_id):
        self.addCell(permission_id, principal_id, True)

    def denyPermissionToPrincipal(self, permission_id, principal_id):
        self.addCell(permission_id, principal_id, False)

    unsetPermissionForPrincipal = PPMSecurityMap.delCell
    getPrincipalsForPermission = PPMSecurityMap.getRow
    getPermissionsForPrincipal = PPMSecurityMap.getCol

    def getSetting(self, permission_id, principal_id, default=None):
        return self.queryCell(permission_id, principal_id, default)

    getPrincipalsAndPermissions = PPMSecurityMap.getAllCells


@implementer(IPrincipalRoleManager)
class PrincipalRoleManager(PRMSecurityMap):
    """Mappings between principals and roles."""


    def assignRoleToPrincipal(self, role_id, principal_id):
        self.addCell(role_id, principal_id, True)

    def removeRoleFromPrincipal(self, role_id, principal_id):
        self.addCell(role_id, principal_id, False)

    unsetRoleForPrincipal = PRMSecurityMap.delCell
    getPrincipalsForRole = PRMSecurityMap.getRow
    getRolesForPrincipal = PRMSecurityMap.getCol

    def getSetting(self, role_id, principal_id):
        return self.queryCell(role_id, principal_id, default=None)

    getPrincipalsAndRoles = PRMSecurityMap.getAllCells


@implementer(IRolePermissionManager)
class RolePermissionManager(RPMSecurityMap):
    """Provide adapter that manages role permission data in an object attribute
    """


    def grantPermissionToRole(self, permission_id, role_id):
        self.addCell(permission_id, role_id, True)

    def denyPermissionToRole(self, permission_id, role_id):
        self.addCell(permission_id, role_id, False)

    unsetPermissionFromRole = RPMSecurityMap.delCell
    getRolesForPermission = RPMSecurityMap.getRow
    getPermissionsForRole = RPMSecurityMap.getCol
    getRolesAndPermissions = RPMSecurityMap.getAllCells

    def getSetting(self, permission_id, role_id):
        return self.queryCell(permission_id, role_id, default=None)


@implementer(IGrantInfo)
class GrantInfoAdapter(object):
    """Grant info adapter.

    Right now we do not provide special pages whihc use the GrantInfo adapter.
    This means we return Allow, Deny and Unset as grant info. Later we will
    probably implement custom grant info pages and use True, False and None as
    grant info values.
    """

    zope.component.adapts(interfaces.ISecurityAware)

    prinper = {}
    prinrole = {}
    permrole = {}

    def __init__(self, context):
        self.context = removeSecurityProxy(context)
        self.prinper = self.context.ppmcol
        self.prinrole = self.context.prmcol
        self.permrole = self.context.rpmrow

    def __bool__(self):
        return bool(self.prinper or self.prinrole or self.permrole)

    def principalPermissionGrant(self, principal, permission):
        """This is the only method which uses Allo, Deny and Unset.

        All other adapter and the security policy use True, False and None.
        """
        prinper = self.prinper.get(principal)
        if prinper:
            grant = prinper.get(permission, Unset)
            if grant is False:
                return Deny
            elif grant is True:
                return Allow
        return Unset

    def getRolesForPermission(self, permission):
        permrole = self.permrole.get(permission)
        if permrole:
            return list(permrole.items())
        return ()

    def getRolesForPrincipal(self, principal):
        prinrole = self.prinrole.get(principal)
        if prinrole:
            return list(prinrole.items())
        return ()




###############################################################################
#
# Copyright (c) 2011 Projekt01 GmbH and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
###############################################################################
"""
$Id:$
"""
from __future__ import absolute_import
from builtins import str
from builtins import object
__docformat__ = "reStructuredText"

import datetime

import bson.son

import zope.schema
from zope.security.proxy import removeSecurityProxy

import m01.mongo.schema
import m01.mongo.util
import m01.mongo.tm

_marker = object()

from m01.mongo import LOCAL


class MongoFieldProperty(object):
    """Computed attributes based on schema fields

    Mongo field property provide default values, data validation and error
    messages based on data found in field meta-data.

    Note that MongoFieldProperty cannot be used with slots. They can only
    be used for attributes stored in instance dictionaries.

    Use this field property for any kind of values except for store MongoObject
    as values on persistent zope objects.

    """

    def __init__(self, field, name=None):
        if name is None:
            name = field.__name__
        self.__field = field
        self.__name = name
        self.__type = field._type
        # preconditions
        if isinstance(field, m01.mongo.schema.MongoBinary):
            raise TypeError(
                "Use MongoBinaryProperty for MongoBinary field %s" % self.__name)
        if isinstance(field, zope.schema.Date):
            raise TypeError(
                "Use MongoDateProperty for Date field %s" % self.__name)
        if isinstance(field, (zope.schema.List, zope.schema.Tuple)):
            raise TypeError(
                "Use MongoList for List or Tuple field %s" % self.__name)
        # setup data factory
        if isinstance(field, m01.mongo.schema.MongoList):
            if isinstance(field.value_type, zope.schema.Object):
                self.__data = m01.mongo.util.MongoItemsData
            else:
                self.__data = m01.mongo.util.MongoListData

    def applyChanges(self, inst, value):
        """Apply changes if they don't compare"""
        current = self.__get__(inst, None)
        if current != value:
            # apply changed value
            inst.__dict__[self.__name] = value
            if inst._m_initialized:
                # mark as changed, but only if instance is initialized
                inst._m_changed = True

    def __get__(self, inst, klass):
        if inst is None:
            return self
        value = inst.__dict__.get(self.__name, _marker)
        if value is _marker:
            field = self.__field.bind(inst)
            value = getattr(field, 'default', _marker)
            if value is _marker:
                raise AttributeError(self.__name)
        if isinstance(value, (tuple, list)):
            # we could get a default value which is an empty list. If so,
            # replace the empty default list value with our item data storage
            # which we setup in our __init_ method.
            value = self.__data(value)
            value.__parent__ = inst
            value.__name__ = self.__name
            inst.__dict__[self.__name] = value
        # take care with this value, it could be a global referenced schema
        # default value. See m01.mongo.util.MongoListData.__init__ for info
        return value

# XXX: why do we not convert any value to it's field._type ?
# e.g.
# if value is not None and not isinstance(value, self.__type):
    def __set__(self, inst, value):
        # first locate our value if there is a __parent__ attr
        if hasattr(value, '__parent__'):
            value.__parent__ = inst
        if isinstance(value, (tuple, list)):
            # convert list values. Take care, the value could be a global
            # referenced schema default value. But our __data list can handle
            # that.
            value = self.__data(value)
            value.__parent__ = inst
            value.__name__ = self.__name
        elif self.__type == str and value is not None:
            # we get a unicode value from MongoDB, convert the value to str
            value = self.__type(value)
        # now validate
        field = self.__field.bind(inst)
        field.validate(value)
        if field.readonly and self.__name in inst.__dict__:
            raise ValueError(self.__name, 'field is readonly')
        # apply changes
        self.applyChanges(inst, value)

    def __delete__(self, inst):
        value = inst.__dict__.get(self.__name, _marker)
        if value is not _marker:
            del inst.__dict__[self.__name]
            inst._m_changed = True

    def __getattr__(self, name):
        return getattr(self.__field, name)


class MongoBinaryProperty(object):
    """Computed attributes based on schema fields

    Mongo field property which will convert bytes input to bson Binary data

    """

    def __init__(self, field, name=None):
        if name is None:
            name = field.__name__
        self.__field = field
        self.__name = name
        self.__type = field._type
        # preconditions
        if not isinstance(field, zope.schema.Bytes):
            raise TypeError(
                "MongoBinaryProperty can only convert zope.schema.Bytes %s",
                    self.__name)

    def applyChanges(self, inst, value):
        """Apply changes if they don't compare"""
        current = self.__get__(inst, None)
        if current != value:
            # apply changed value
            inst.__dict__[self.__name] = value
            if inst._m_initialized:
                # mark as changed, but only if instance is initialized
                inst._m_changed = True

    def __get__(self, inst, klass):
        if inst is None:
            return self
        value = inst.__dict__.get(self.__name, _marker)
        if value is _marker:
            field = self.__field.bind(inst)
            value = getattr(field, 'default', _marker)
            if value is _marker:
                raise AttributeError(self.__name)
        if value is not None and not isinstance(value, self.__type):
            value = self.__type(value)
        return value

    def __set__(self, inst, value):
        if value is not None and not isinstance(value, self.__type):
            value = self.__type(value)
        # validate
        field = self.__field.bind(inst)
        field.validate(value)
        if field.readonly and self.__name in inst.__dict__:
            raise ValueError(self.__name, 'field is readonly')
        # apply changes
        self.applyChanges(inst, value)

    def __delete__(self, inst):
        value = inst.__dict__.get(self.__name, _marker)
        if value is not _marker:
            del inst.__dict__[self.__name]
            inst._m_changed = True

    def __getattr__(self, name):
        return getattr(self.__field, name)


class MongoDateProperty(object):
    """Computed attributes based on schema fields

    Mongo field property which will convert a given Date to ordinal and back.

    Note, we need to do this beacuse BSON and pymongo do not support date
    objects. There is also a restriction in datetime conversion.

    Note: we need to do this because pymongo uses calendar.timegm() and
    datetime.utcfromtimestamp for datetime object. This won't work for
    datetime object outside 1970 - 2038.
    """

    def __init__(self, field, name=None):
        if name is None:
            name = field.__name__
        self.__field = field
        self.__name = name
        if not isinstance(field, m01.mongo.schema.MongoDate):
            raise TypeError(
                "Only use MongoDateProperty for MongoDate field")

    def applyChanges(self, inst, value):
        """Apply changes if they don't compare"""
        current = self.__get__(inst, None)
        if current != value:
            # apply changed value
            inst.__dict__[self.__name] = value
            if inst._m_initialized:
                # mark as changed, but only if instance is initialized
                inst._m_changed = True

    def __get__(self, inst, klass):
        # ordinal to date conversion
        if inst is None:
            return self
        value = inst.__dict__.get(self.__name, _marker)
        if value is _marker:
            field = self.__field.bind(inst)
            value = getattr(field, 'default', _marker)
            if value is _marker:
                raise AttributeError(self.__name)
        # convert to date
        if value is not None:
            return datetime.date.fromordinal(value)
        return value

    def __set__(self, inst, value):
        # date to ordinal conversion
        if isinstance(value, int):
            value = datetime.date.fromordinal(value)
        field = self.__field.bind(inst)
        field.validate(value)
        if field.readonly and self.__name in inst.__dict__:
            raise ValueError(self.__name, 'field is readonly')
        if value is not None:
            value = value.toordinal()
        # apply changes
        self.applyChanges(inst, value)

    def __delete__(self, inst):
        value = inst.__dict__.get(self.__name, _marker)
        if value is not _marker:
            del inst.__dict__[self.__name]
            inst._m_changed = True

    def __getattr__(self, name):
        return getattr(self.__field, name)


class MongoGeoLocationProperty(object):
    """Computed GeoLocation based on given data

    Note: We will compare the values an prevent write access for not changed
    items. This means the MongoGeoLocationProperty can only save Geolocation
    items with lon, lat values and not enhanced implementations with additional
    values.
    """

    def __init__(self, field, name=None, geoLocationFactory=None):
        if name is None:
            name = field.__name__
        self.__field = field
        self.__name = name
        if geoLocationFactory is None:
            # prevent recursive import
            from m01.mongo.geo import GeoLocation
            geoLocationFactory = GeoLocation
        self.__factory = geoLocationFactory
        # preconditions
        if not isinstance(field, zope.schema.Object):
            raise TypeError(
                "Use zope.schema.Object for GeoLocation field %s" % self.__name)

    def __get__(self, inst, klass):
        if inst is None:
            return self
        value = inst.__dict__.get(self.__name, _marker)
        if value is _marker:
            field = self.__field.bind(inst)
            value = getattr(field, 'default', _marker)
            if value is _marker:
                raise AttributeError(self.__name)
        return value

    def __set__(self, inst, value):
        # allowed values are a dict with {'lon', <lon>, 'lat': <lat>} or an
        # IGeoLocation instance or even a list with lon, lat values
        if isinstance(value, (dict, list)):
            # convert dict data
            value = self.__factory(value)
        if value is not None:
            # never store proxied objects
            value = removeSecurityProxy(value)
            # locate them
            value.__name__ = str(self.__name)
            value.__parent__ = inst
        # now validate
        field = self.__field.bind(inst)
        field.validate(value)
        if field.readonly and self.__name in inst.__dict__:
            raise ValueError(self.__name, 'field is readonly')
        # since z3c.form can't simply check object items, we will only save a
        # geo location if the lat, lon will be different then the current
        # stored values. This will prevent useless write access
        cLat = None
        cLon = None
        vLat = None
        vLon = None
        current = inst.__dict__.get(self.__name)
        if current is not None:
            cLat = current.lat
            cLon = current.lon
        if value is not None:
            vLat = value.lat
            vLon = value.lon
        if not (cLat == vLat and cLon == vLon):
            inst.__dict__[self.__name] = value
            if inst._m_initialized:
                # mark as changed
                inst._m_changed = True

    def __delete__(self, inst):
        value = inst.__dict__.get(self.__name, _marker)
        if value is not _marker:
            del inst.__dict__[self.__name]
            inst._m_changed = True

    def __getattr__(self, name):
        return getattr(self.__field, name)


class MongoGeoPointProperty(object):
    """Computed GeoPoint based on given data

    Note: We will compare the values an prevent write access for not changed
    items. This means the MongoGeoPointProperty can only save GeoPoint
    items with lon, lat values and not enhanced implementations with additional
    values.
    """

    def __init__(self, field, name=None, geoLocationFactory=None):
        if name is None:
            name = field.__name__
        self.__field = field
        self.__name = name
        if geoLocationFactory is None:
            # prevent recursive import
            from m01.mongo.geo import GeoPoint
            geoLocationFactory = GeoPoint
        self.__factory = geoLocationFactory
        # preconditions
        if not isinstance(field, zope.schema.Object):
            raise TypeError(
                "Use zope.schema.Object for GeoPoint field %s" % self.__name)

    def __get__(self, inst, klass):
        if inst is None:
            return self
        value = inst.__dict__.get(self.__name, _marker)
        if value is _marker:
            field = self.__field.bind(inst)
            value = getattr(field, 'default', _marker)
            if value is _marker:
                raise AttributeError(self.__name)
        return value

    def __set__(self, inst, value):
        # allowed values are a dict with {'lon', <lon>, 'lat': <lat>} or an
        # IGeoPoint instance or even a list with lon, lat values
        if isinstance(value, (dict, bson.son.SON, list)):
            # convert dict data
            value = self.__factory(value)
        if value is not None:
            # never store proxied objects
            value = removeSecurityProxy(value)
            # locate them
            value.__name__ = str(self.__name)
            value.__parent__ = inst
        # now validate
        field = self.__field.bind(inst)
        field.validate(value)
        if field.readonly and self.__name in inst.__dict__:
            raise ValueError(self.__name, 'field is readonly')
        # since z3c.form can't simply check object items, we will only save a
        # geo location if the lat, lon will be different then the current
        # stored values. This will prevent useless write access
        cLat = None
        cLon = None
        vLat = None
        vLon = None
        current = inst.__dict__.get(self.__name)
        if current is not None:
            cLat = current.lat
            cLon = current.lon
        if value is not None:
            vLat = value.lat
            vLon = value.lon
        if not (cLat == vLat and cLon == vLon):
            inst.__dict__[self.__name] = value
            if inst._m_initialized:
                # mark as changed
                inst._m_changed = True

    def __delete__(self, inst):
        value = inst.__dict__.get(self.__name, _marker)
        if value is not _marker:
            del inst.__dict__[self.__name]
            inst._m_changed = True

    def __getattr__(self, name):
        return getattr(self.__field, name)


class MongoSubObjectProperty(object):
    """Computed attributes based on zope.schema.Object

    The field property will convert the given data to an object using the
    given converter method or class. If None data is given for this field,
    the field property does not create and add an object.

    The field property get method will create a new object if None is stored
    and a factor class or method is given.

    Note: the MongoSubObjectProperty will store the object in the instance
    using this property and the MongoOjectProperty will store object in another
    collection.

    """

    def __init__(self, field, converter, factory=None, name=None):
        if name is None:
            name = field.__name__
        self.__field = field
        self.__converter = converter
        self.__factory = factory
        self.__name = name
        self.__required = field.required
        if field.readonly:
            raise TypeError("Can't use readonly fields")
        # preconditions
        if not isinstance(field, zope.schema.Object):
            raise TypeError(
                "Use zope.schema.Object for GeoLocation field %s" % self.__name)

    def __get__(self, inst, klass):
        if inst is None:
            return self
        obj = inst.__dict__.get(self.__name)
        if obj is None:
            field = self.__field.bind(inst)
            obj = getattr(field, 'default', None)
        if obj is None and self.__factory is not None:
            obj = self.__factory({
                '__name__': self.__name,
                '__parent___': inst,
                })
            inst.__dict__[self.__name] = obj
        return obj

    def __set__(self, inst, data):
        # convert the value if a converter is given
        if data is None:
            # convert None value to empty dict
            data = {}
        if isinstance(data, dict):
            # data given from mongo as bson otherwise we will add an object
            data['__name__'] = self.__name
            data['__parent__'] = inst
            obj = self.__converter(data)
        else:
            obj = data
            if obj.__parent__ is None:
                obj.__parent__ = inst
            elif obj.__parent__ is not inst:
                raise ValueError(
                    "Wrong __parent__ given, __parent__ must be %r" % inst,
                    obj.__parent__)
            # set __name__ if not given
            if obj.__name__ is None:
                obj.__name__ = str(self.__name)
        # validate the obj
        field = self.__field.bind(inst)
        field.validate(obj)
        if field.readonly and self.__name in inst.__dict__:
            raise ValueError(self.__name, 'field is readonly')
        # apply changed obj
        inst.__dict__[self.__name] = obj
        if inst._m_initialized:
            # mark as changed, but only if instance is initialized
            inst._m_changed = True

    def __getattr__(self, name):
        return getattr(self.__field, name)


class MongoObjectProperty(object):
    """Computed MongoObject based on schema fields and given configuration
    using an unique _moid attribute as __parent__ reference and an attribute
    name for build a unique _oid.

    MongoObjectProperty provides default values, data validation and error
    messages based on data found in field meta-data.

    The MongoObjectProperty can read and write a MongoObject from and to MongoDB.
    The access to the MongoDB is lazy which means the object get only loaded if
    we access them. A transaction data manager is used for write data back to
    the MongoDB on transaction commit.

    Use this field property only for store MongoObject objects as attribute
    values. Use the default MongoFieldProperty and MongoSubItem objects if you
    like to store objects as list or dict items.

    This property is used if you need to set a MongoObject as an attribute
    on an object. Such MongoObject attributes get stored in a MongoDB
    collection defined by the getCollection method.

    This property can get used on persistent zope object or on non persistent
    MongoItem based objects. If you use a MongoObjectProperty in a MongoItem,
    you don't have to setup or dump such MongoObject attribute values. They
    are full independent from their MongoAwareObject.

    MongoObject provide everything a normal MongoItem can provide. Which means
    you can use MongoSubItem, converters, etc.

    """

    def __init__(self, field, factory, getCollection=None):
        self.__factory = factory
        if getCollection is None:
            self.__getCollection = factory.getCollection
        else:
            self.__getCollection = getCollection
        self.__field = field
        self.__name = field.__name__
        if not isinstance(field, zope.schema.Object):
            raise TypeError(
                "Can only use MongoObjectProperty for zope.schema.Object field")

    def getMongoObject(self, inst):
        _oid = u'%s:%s' % (inst._moid, self.__name)
        value = LOCAL.__dict__.get(_oid, _marker)
        if value is _marker:
            collection = self.__getCollection(inst)
            data = collection.find_one({'_oid':_oid})
            if data is not None and not data.get('removed'):
                data['__parent__'] = inst
                if data.get('_field') is None:
                    # apply _field value (new since 0.6.0)
                    # previous version provided the field.__name__ as __name__
                    # this didn't allow to use a custom __name__ and traverse
                    # the item from differetn containers. The new pattern stores
                    # the field name where the MongoObject is stored in as
                    # _field value. This allows to use different values for
                    # __name__ and _field.
                    data['_field'] = str(self.__name)
                    # provide previous __name__ as _field value
                assert data['_field'] == self.__name
                value = self.__factory(data)
                # cache mongo object
                LOCAL.__dict__[_oid] = value
                # setup transaction handling
                m01.mongo.tm.ensureMongoTransaction(value)
        elif value.removed:
            value = _marker
        return value

    def __get__(self, inst, klass):
        if inst is None:
            return self
        value = self.getMongoObject(inst)
        if value is _marker:
            # find default value
            field = self.__field.bind(inst)
            value = getattr(field, 'default', _marker)
            if value is _marker:
                raise AttributeError(self.__name)
        return value

    def __set__(self, inst, value):
        # Note, we do not need to remove existing items, they get overriden
        # with the new one because a new MongoObject provides the same _oid.
        if not hasattr(inst, '_moid'):
            raise ValueError("MongoObject __parent__ doens't provide an _moid")

        # remove old item
        old = self.getMongoObject(inst)
        if old is not _marker:
            # mark as removed and _m_changed which forces that the object get
            # removed in voteCommit call. Note, this part get only called if
            # we set another and not a new or initial value from mongodb.
            old.removed = True
            # setup transaction and append the old MongoObject before we add
            # a new one. This will ensure that we remove the item from MongoDB
            # before we add the new one. Note: both use the same _oid
            m01.mongo.tm.ensureMongoTransaction(value)
        if value is None:
            # bind and validate, the above code marks an old object as removed
            # if there was any
            field = self.__field.bind(inst)
            field.validate(value)
            return

        # set or compare __parent__ which is required for _oid
        if value.__parent__ is None:
            value.__parent__ = inst
        elif value.__parent__ is not inst:
            raise ValueError(
                "Wrong __parent__ given, __parent__ must be %r" % inst,
                value.__parent__)

        # set __name__ if not given
        if value.__name__ is None:
            value.__name__ = str(self.__name)

        # set or compare _field which is required for _oid
        if value._field is None:
            value._field = str(self.__name)
        elif value._field != self.__name:
            raise ValueError(
                "Wrong _field given, _field must be %r" % self.__name,
                value._field)

        # now we can get and approve the _oid
        _oid = u'%s:%s' % (inst._moid, self.__name)
        if _oid != value._oid:
            raise ValueError("MongoObject _oid %r does not compare", value._oid)

        # bind and validate
        field = self.__field.bind(inst)
        field.validate(value)
        # validate readonly
        if field.readonly and self.__getCollection(inst).find_one(
            {'_oid': value._oid}.count()):
            raise ValueError(self.__name, 'field is readonly')
        # setup transaction and append the MongoObject
        m01.mongo.tm.ensureMongoTransaction(value)
        # cache mongo object
        LOCAL.__dict__[_oid] = value
        if inst._m_initialized:
            # if __parent__ is initialized mark the object as changed, which
            # will mark the __parent__
            value._m_changed = True

    def __delete__(self, inst):
        # this clever concept will force that the MongoObject get removed on
        # transaction commit
        value = self.getMongoObject(inst)
        if value is not _marker:
            # mark as removed and _m_changed which forces that the object get
            # removed in voteCommit call
            value.removed = True

    def __getattr__(self, name):
        return getattr(self.__field, name)



###############################################################################
#
# Copyright (c) 2011 Projekt01 GmbH and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
###############################################################################
"""
$Id:$
"""
from __future__ import absolute_import
from builtins import object
from zope.interface import implementer
__docformat__ = "reStructuredText"

import transaction
from transaction.interfaces import IDataManager

import zope.interface

from m01.mongo import LOCAL


@implementer(IDataManager)
class MongoTransactionDataManager(object):
    """Transaction Data manager for mongo object.

    The transaction data manager is responsible for add, modify and remove a
    mongo object in a mongo database and will ensure cleanup and error handling.

    This transaction manager should be generic since he delegates operations
    to the IMongoStorage.
    """


    def __init__(self, obj, tm, transaction):
        self.transaction_manager = tm # useless but defined by interface
        self.transaction = transaction
        self.objects = []
        self.objects.append(obj)
        self._prepared = False

    def append(self, obj):
        if obj not in self.objects:
            self.objects.append(obj)

    def tpc_begin(self, transaction):
        """Begin commit of a transaction, starting the two-phase commit."""
        if self._prepared:
            raise TypeError('Already prepared')
        self._checkTransaction(transaction)
        self._prepared = True
        self.transaction = transaction
        # before we start processing, remove our thread local reference
        delattr(LOCAL, 'MongoTransactionDataManager')
        for obj in self.objects:
            obj.begin()

    def commit(self, transaction):
        """Delegate the commit call to the object."""
        if not self._prepared:
            raise TypeError('Not prepared to commit')
        self._checkTransaction(transaction)
        self.transaction = None
        self._prepared = False
        for obj in self.objects:
            obj.commit()

    def abort(self, transaction):
        """Abort a transaction and forget all changes.

        This call get dispatched to tpc_abort since abort get not always called
        if we abort a transaction.
        """
        self.tpc_abort(transaction)

    def tpc_finish(self, transaction):
        """Indicate confirmation that the transaction is done.

         This should never fail.
         """
        for obj in self.objects:
            obj.finish()
        # remove objects which will prevent calling this twice
        self.objects = []

    def tpc_vote(self, transaction):
        """Verify that a data manager can commit the transaction."""
        self._checkTransaction(transaction)
        for obj in self.objects:
            obj.vote()

    def tpc_abort(self, transaction):
        """Abort a transaction. This should never fail.

        This method can get called more then once because we only abort on the
        first call. This is important if since abort() call can get skipped if
        an error happens before this data manager get commited.
        """
        self._checkTransaction(transaction)
        if self.transaction is not None:
            self.transaction = None
        self._prepared = False
        for obj in self.objects:
            obj.abort()
        # remove objects which will prevent calling this twice
        self.objects = []

    def sortKey(self):
        # use an order key which (hopefully) forces to order as last item.
        # Remember, python orders a list like:
        # [None, 42, time.time(), 123456789123456789L, 'AA', 'BB', 'aa', 'bb']
        return 'zzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzz'

    # helpers
    def _checkTransaction(self, transaction):
        """Check for a valid transaction."""
        if (self.transaction is not None and
            self.transaction is not transaction):
            raise TypeError("Transaction missmatch", transaction,
                self.transaction)


def ensureMongoTransaction(obj):
    """Observes mongo items

    Threading local provides a reference to our shared transaction data manager.
    This will ensure that we never use more then one transaction data manager
    if more then one storage or object get used.
    """
    tm = transaction.manager
    txn = tm.get()
    tdm = getattr(LOCAL, 'MongoTransactionDataManager', None)
    if tdm is None or tdm.transaction != txn:
        tdm = MongoTransactionDataManager(obj, tm, txn)
        setattr(LOCAL, 'MongoTransactionDataManager', tdm)
        txn.join(tdm)
    else:
        # this doesn't add the same object twice, see the append method. But
        # you allways have to make sure that you never ever use 2 different
        # instances representing the same logical object or if so you must make
        # sure that this objects don't use the same cache key. You will get
        # into trouble if 2 objects represent the same logical item using the
        # same cache key  because the first transaction begin will move the
        # items from the insert cache to the transaction cache and the second
        # will clear the cache before the first calls vote which whould commit
        # the items to the mongodb.
        # The default MongoMappingBase cache key prevents this by using a
        # unique cache key including id(self).
        tdm.append(obj)
    return tdm



###############################################################################
#
# Copyright (c) 2011 Projekt01 GmbH and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
###############################################################################
"""
$Id:$
"""
from __future__ import absolute_import
from builtins import str
__docformat__ = "reStructuredText"

import datetime
import bson.objectid

import zope.interface
import zope.location.interfaces

import m01.mongo
from m01.mongo import UTC
from m01.mongo import LOCAL
from m01.mongo import interfaces
from m01.mongo import base
from m01.mongo.fieldproperty import MongoFieldProperty
from zope.interface import implementer

_marker = object()


class MongoSubItem(base.SetupConvertDump):
    """Mongo sub item base class usable for MongoList schema items.

    A mongo sub item get used in a MongoItems (list) and is a part off a
    MongoItem which is stored in a MongoDB. Such a mongo sub item provide an
    own object _id reference which is stored in the sub items container like
    we use with MongoItemsData by default. As you can see in MongoFieldProperty,
    we need a MongoList schema for setup such a MongoItemsData container.

    """

    _id = None
    _type = None
    _m_changed_value = None

    # built in skip and dump names
    _skipNames = ['__name__']
    _dumpNames = ['_id', '_type', 'created', 'modified']

    # customize this in your implementation
    skipNames = []
    dumpNames = []

    created = MongoFieldProperty(interfaces.IMongoSubItem['created'])
    modified = MongoFieldProperty(interfaces.IMongoSubItem['modified'])

    def __init__(self, data):
        """Initialize a MongoSubItem with given data."""
        # set given or new _id
        _id = data.pop('_id', _marker)
        if _id is _marker:
            _id = bson.objectid.ObjectId()
        self.__dict__['_id'] = _id

        # set given or None __parent__
        __parent__ = data.pop('__parent__', None)
        if __parent__ is not None:
            self.__parent__ = __parent__

        # set given or new _type
        _type = data.pop('_type', self.__class__.__name__)
        if _type != self.__class__.__name__:
            raise TypeError("Wrong mongo item _type used")
        self.__dict__['_type'] = str(_type)

        # set given or new created datetime
        created = data.pop('created', _marker)
        if created is _marker:
            created = m01.mongo.now(UTC)
        self.__dict__['created'] = created

        # update object with given key/value data
        self.setup(data)

        # it is very important to set _m_changed_value to None, otherwise each
        # read access will end in a write access.
        self._m_changed_value = None

    @property
    def __name__(self):
        return str(self._id)

    def _m_changed():
        def fget(self):
            return self._m_changed_value
        def fset(self, value):
            if value == True:
                self._m_changed_value = value
                if self.__parent__ is not None:
                    self.__parent__._m_changed = value
            elif value != True:
                raise ValueError("Can only dispatch True to __parent__")
        return property(fget, fset)

    _m_changed = _m_changed()

# TODO: this is not the best contept. Use a comit marker and implement this
#       in Dump mixin:
#    def dump(self, dumpNames=None, dictify=False, commit=False):
#        # prevent write access
#        if self._m_changed and commit:
#            self.modified = m01.mongo.now(UTC)
#        return super(MongoSubItem, self).dump(dumpNames, commit)
# or even better use a private attr for modified e.g. _m_modified and write to
# this attr in _m_changed. This would write _m_modified on each _m_changed but
# it would only write the new changed date if we commit the parent obj.
# probably use a _m_modified_original and keep a reference to it and reset the
# _m_modified if we set _m_changed to None again which also could happen
    def dump(self, dumpNames=None, dictify=False):
        if self._m_changed and (dumpNames is None or 'modified' in dumpNames):
            self.modified = m01.mongo.now(UTC)
        return super(MongoSubItem, self).dump(dumpNames)


@implementer(interfaces.IMongoContainerItem, interfaces.IMongoParentAware, zope.location.interfaces.ILocation)
class MongoContainerItem(base.MongoItemBase):
    """Simple mongo container item.

    Implement your own IMongoContainerItem with the attributes you need.
    """

    # validate __name__ and observe to set _m_changed
    __name__ = MongoFieldProperty(zope.location.interfaces.ILocation['__name__'])


@implementer(interfaces.IMongoStorageItem, interfaces.IMongoParentAware, zope.location.interfaces.ILocation)
class MongoStorageItem(base.MongoItemBase):
    """Simple mongo storage item.

    This MongoItem will use the mongo ObjectId as the __name__. This means
    you can't set an own __name__ value for this object.

    Implement your own IMongoStorageItem with the attributes you need.
    """

    _skipNames = ['__name__']

    @property
    def __name__(self):
        return str(self._id)


@implementer(interfaces.ISecureMongoContainerItem, interfaces.IMongoParentAware, zope.location.interfaces.ILocation)
class SecureMongoContainerItem(base.SecureMongoItemBase):
    """Security aware MongoContainerItem."""

    # validate __name__ and observe to set _m_changed
    __name__ = MongoFieldProperty(zope.location.interfaces.ILocation['__name__'])


@implementer(interfaces.ISecureMongoStorageItem, interfaces.IMongoParentAware, zope.location.interfaces.ILocation)
class SecureMongoStorageItem(base.SecureMongoItemBase):
    """Security aware MongoStorageItem."""

    _skipNames = ['__name__']

    @property
    def __name__(self):
        return str(self._id)


@implementer(interfaces.IMongoSubObject, zope.location.interfaces.ILocation)
class MongoSubObject(base.MongoSubObjectBase):
    """Mongo sub Object class"""


@implementer(interfaces.IMongoObject, zope.location.interfaces.ILocation)
class MongoObject(base.MongoItemBase):
    """Mongo object used for store as object value on persistent or non
    persistent objects.

    This mongo object is used as object value and must provide an own
    collection method. This means such an object get not stored in the same
    collection as the __parent__ object .The __parent__ could be a mongo item,
    sub item or container etc.

    The attribute name is used as the mongo object _field. This _field
    is together with the mongo object parent id (_moid) used as unique id.
    This unique id is exposed as _oid value. Also see MongoObjectProperty

    The __parent__ object which contains such MongoObject items needs to
    provide an _moid attribute with a unique object id. This _moid is used as a
    part of unique mongo id (_oid). You can use the p01.oid package which
    provides such a unique id for peristent __parent__ objects if you mix ZODB
    and ZODB less databases or you can use the MongoObjectAware class for non
    persistent items. If you use the p01.oid package, you need to expose the
    oid as _moid in your implementation. See how we did it in MongoObjectAware

    Note: If you enhance your own MongoObject __init__ method, don't forget
    to set_m_changed to None. Otherwise it will end in write access on
    each read.

    Changes:
    Since version 0.6.0 we will use the _field value as the MongoObject part
    for _oid. Previous version where using the __name__ instead. For older
    MongoObject items stored in mongodb we will support boths. This means if
    a _field valu is given we use them if not we use the __name__.

    New items must set the __name__ explict if they need one. Note, a __name__
    allows to traverser the MongoObject item within a different container
    by it's __name__. This is the reason why we changed the implementation.
    Now it's possible to store the MongoObject as property valu in a parent
    item and at the same time to implement a container or storage and traverse
    them as container item. This is a core pattern for gridfs files implemented
    in m01.grid (see m01.grid.item.FileObject)

    Note:
    The MongoObject implementation can't get mixed into one class with
    MongoContainerItem, MongoStorageItem or MongoItem beacause of the
    transaction commit process API.

    We currently only use the MongoObject items with read container which only
    allow to traverse to the MongoObject items. We have not checked if it is
    possible to implement a container or storage which is able to write changes
    in MongoObject items to the collection like we do with MongoContainerItem
    or MongoStorageItem.

    The current API looks like it should be possible to implement a MongoObject
    as item. If we do so in the future, we probably offer a IMongoObjectStorage
    and MongoObjectContainer. The big question is does such a container
    implementation support to store items in different collections?
    """

    # write concern markers (empty dict means connection settings get used)
    _m_insert_write_concern = {}
    _m_update_write_concern = {}
    _m_remove_write_concern = {}

    _m_independent = False

    # marker for remove handling
    _field = MongoFieldProperty(interfaces.IMongoObject['_field'])
    removed = MongoFieldProperty(interfaces.IMongoObject['removed'])

    _skipNames = ['_oid']
    _dumpNames = ['_id', '_oid', '__name__', '_type', '_version',
                  '_field',
                  'created', 'modified', 'removed']

    @classmethod
    def getCollection(cls, parent):
        """Note, this method must be a classmethod and return the collection."""
        raise NotImplementedError(
            "Subclass must implement classmethod getCollection")

    @property
    def collection(self):
        return self.getCollection(self.__parent__)

    @property
    def _oid(self):
        try:
            # BBB before 0.6.0, never remove
            # the _field value is new since 0.6.0. Previous version where using
            # the __name__ instead. Support both
            _field = self._field and self._field or self.__name__
            return u'%s:%s' % (self.__parent__._moid, _field)
        except AttributeError as e:
            if self.__parent__ is None:
                raise AttributeError("Missing __parent__ object")
            else:
                raise AttributeError(
                    '__parent__ "%s" does not provide an _moid' % (
                        self.__parent__))

    # MongoDB handler methods
    def doInsert(self):
        self.modified = m01.mongo.now(UTC)
        self.collection.insert_one(self.dump())

    def doUpdate(self, upsert=False):
        self.modified = m01.mongo.now(UTC)
        data = self.dump()
        # skip _id with $set, OperationFailure: Mod on _id not allowed
        data.pop('_id', None)
        self.collection.update_one({'_id': self._id}, {'$set': data},
            upsert=upsert)

    def doRemove(self):
        # never call remove without an _id
        if self._id is None:
            raise ValueError("Empty _id given, this would remove all objects")
        self.collection.delete_one({'_id': self._id})
        if self._oid in LOCAL.__dict__:
            del LOCAL.__dict__[self._oid]

    def begin(self):
        """Beginn commit a transaction handling."""
        # compare not independent items by it's _id and version number
        if self._version > 0 and not self._m_independent \
            and not self.collection.find_one(
            {'_id':self._id, '_version':self._version}):
            raise ValueError("MongoItem version does not compare")

    # transaction management methods
    def commit(self):
        """Start commit transaction handling."""
        pass

    def vote(self):
        """Process MongoDB insert and update calls.

        Note, if something fails, this could leave inconsistent data in MongoDB.
        """
        if self._m_changed:
            # remove marked item
            if self.removed:
                self.doRemove(**self._m_remove_write_concern)
            else:
                # handle not removed item
                self._version += 1
                if self._version == 1:
                    # insert new item
                    self.doInsert(**self._m_insert_write_concern)
                else:
                    # update existing items
                    self.doUpdate(**self._m_update_write_concern)

    def abort(self):
        """Abort transaction handling."""
        pass

    def finish(self):
        """Finish transaction commit."""
        if self._oid in LOCAL.__dict__:
            del LOCAL.__dict__[self._oid]



###############################################################################
#
# Copyright (c) 2011 Projekt01 GmbH and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
###############################################################################
"""
$Id:$
"""
from __future__ import division
from __future__ import absolute_import
from zope.interface import implementer
from future import standard_library
import six
standard_library.install_aliases()
from builtins import str
from six import string_types
from past.utils import old_div
from future.utils import PY2
from future.types import newstr

import _thread
import datetime

import pymongo
import pymongo.errors
import bson.objectid
import bson.son

import transaction
import zope.interface
import zope.location.location
from zope.security.proxy import removeSecurityProxy

import m01.mongo
from m01.mongo import UTC
from m01.mongo import LOCAL
from m01.mongo import interfaces
from m01.mongo import util
from m01.mongo.fieldproperty import MongoFieldProperty
from m01.mongo.tm import ensureMongoTransaction

_marker = object()


class MongoObjectAware(object):
    """Offers _moid (mongo object id) referencable for IMongoObject as _oid"""

    # mongo id, built in attrs
    _id = None

    # mongo object _id (_moid) used as a _oid reference in IMongoObject, by
    # default None. This additional _moid (_id) to _oid mapping offers a hook
    # for custom _moid concepts independent from other _id refences like _mpid.
    # See IMongoItemAware for more info.
    @property
    def _moid(self):
        return self._id


class MongoItemAware(object):
    """Offers _mpid (mongo parent id) referencable for IMongoParentAware item"""

    # mongo id, built in attrs
    _id = None

    # mongo _id used as __parent__ reference, by default None. This additional
    # _id to _mpid mapping offers a hook for custom _pid concepts. Which means
    # you can use a similar _mpid in different containers for sharing items.
    # Note: no _mpid means no __parent__ filter and all items in the same
    # collection get shared as long as no other filters get used. See doFind
    # doFindOne etc. methods how we filter by default
    @property
    def _mpid(self):
        if self._id is not None:
            return self._id


class MongoParentAware(object):
    """References IMongoParent item by it's _mpid stored as _pid

    Used for restrict an item to a given __parent__ using a _mpid attribute.

    If no _mpid is given, the items belong to any container which will load
    items without additional filters. See doFind, doFindOne etc. methods for
    more information about filter items.
    """

    # built in attrs
    _pid = None
    _m_parent = None

    def __parent__():
        def fget(self):
            return self._m_parent
        def fset(self, __parent__):
            self._m_parent = __parent__
            # only set _pid the first time
            if self._pid is None:
                try:
                    self._pid = __parent__._mpid
                except AttributeError as e:
                    pass
        return property(fget, fset)

    __parent__ = __parent__()


class SetupConvert(zope.location.location.Location):
    """Mixin class for setup, update and dump item and subitems.

    Note: a converter dict can contain a converter method which returns a
    converted value or a class. A class can be used if we need to create sub
    items values. e.g. embeded mongo array data
    """

    # built in attrs without change observation. Note: this base class does not
    # support _pid (__parent__) reference. See MongoItemBase for _pid support
    __parent__ = None
    __name__ = None
    _m_initialized = False
    _m_changed = None

    # _skipNames defines which attribute/valus will not get stored containd in
    # the datat from mongodb. The _skipNames list will get overriden with the
    # default skipped attribute names from the base class which implements this
    # class. Don't customize this in your custom implementation or if so, make
    # sure you will define every skipped attribute name even those which will
    # get added in future release.
    _skipNames = []

    # customize this in your implementation
    skipNames = [] # attribute names
    converters = {} # attr-name/converter
    defaults = {} # attr-name/value

    def convert(self, key, value):
        """This convert method knows how to handle nested converters.

        A converter can convert attribute values if the attribute
        value is a list of items. You can use a method which knows how to
        handle each value type.

        You can define a converter method which knows to convert all kind
        of items like:

        def toMyItems(value):
            _type = value.get('_type')
            if _type == 'Car':
                return Car(value)
            elif _type == 'House':
                return House(value)
            else:
                return value

        converters = {'myItems': toMyItems}

        IMPORTANT; this converter can NOT convert a nested data structure, it
        is NOT possible to define converters like this:

        e.g.
        def toCar(value):
            return Car(value)

        converters = {'myItems': {'house': toHouse, 'car': toCar}}

        """
        converter = self.converters.get(key)
        if converter is not None:
            if isinstance(value, (list, tuple)):
                # convert list values
                value = [converter(d) for d in value]
            elif isinstance(value, dict):
                # convert mapping values
                if hasattr(value, '__parent__') and value.__parent__ is None:
                    # apply __parent__ if value.__parent__ is None
                    value.__parent__ = self
                value = converter(value)
        return value

    def setup(self, data, dumpOriginalData=False):
        """Setup object with given data key/values. See IMongoHandler"""
        if dumpOriginalData:
            orgData = {'_m_changed': self._m_changed, '_version': self._version}
        else:
            orgData = None

        # setup key/values
        skipNames = self._skipNames + self.skipNames
        for k, v in list(data.items()):
            # skip names if given in data dict
            if k not in skipNames:
                # dump old data
                if orgData is not None:
                    if isinstance(getattr(self, k), (list, tuple)):
                        orgData[k] = [obj.dump() for obj in getattr(self, k, [])]
                    else:
                        orgData[k] = getattr(self, k)
                # convert value
                v = self.convert(k, v)
                # set value
                try:
                    setattr(self, k, v)
                except zope.interface.Invalid as e:
                    pass
                except (AttributeError, ValueError) as e:
                    raise AttributeError(
                        "Can't set value %s %s for attr %s: %r on %r" % (
                            v, type(v), k, str(e), self))

        # this allows to ensure default values or even converter instances
        for k, v in list(self.defaults.items()):
            if getattr(self, k, None) is None:
                # convert value or return an initial empty converter instance
                v = self.convert(k, v)
                setattr(self, k, v)

        # mark the item as initialized. This allows us to implement attribtues
        # with apply decorator which act different for an initialized object
        self._m_initialized = True

        # return original data or None
        return orgData

    def __repr__(self):
        return '<%s %r>' % (self.__class__.__name__, self.__name__)

NO_VALUE_MARKER = object()

class Dump(object):
    """Mixin class for dump item and subitems."""

    # _dumpNames defines which attributes will get stored in mongodb. This
    # list of attribute names will get overriden with default names from the
    # base class which implements this class. Don't customize this list in your
    # custom implementation or if so, make sure you will define every attribute even those which will
    # name which will get added in future release.
    _dumpNames = []
    # customize this in your implementation
    dumpNames = [] # attribute names

    def dump(self, dumpNames=None, dictify=False):
        """Dump the object to json data dict providing a sorted order."""
        if dumpNames is None:
            dumpNames = self._dumpNames + self.dumpNames
        # allways preserve order in mongodb
        data = bson.son.SON()
        to_lower = lambda x: x.lower() if isinstance(x, str) else x.decode('utf-8').lower()
        for k in sorted(dumpNames, key=to_lower):
            # get value from __dict__, this will ensure that we catch values
            # converted by field properties e.g. MongoDateProperty
            v = self.__dict__.get(k, NO_VALUE_MARKER)
            if v is NO_VALUE_MARKER:
                # this will make sure that we catch values which are not stored
                # in our __dict__ like object properties and schema default
                # values
                v = getattr(self, k, NO_VALUE_MARKER)
            # if we found a value dump and/or store it
            if v is not NO_VALUE_MARKER:
                # found a value or default schema value
                if hasattr(v, 'dump'):
                    # dump MongoSubItem, MongoDubObject, MongoItemsData or
                    # MongoListData
                    data[k] = v.dump()
                else:
                    # store simple value including None if None is given
                    data[k] = v
        if dictify:
            data = m01.mongo.dictify(data)
        return data


class SetupConvertDump(SetupConvert, Dump):
    """Mixin class for update items without transaction support"""


@implementer(interfaces.IMongoUpdateMixin)
class MongoUpdateMixin(object):
    """Optional update method mixin class.

    This mixin class offers a method which allows to update a MongoItem without
    a transaction process. You can use this mixin class for update counters or
    other not so important values. Not, this get probably called more then once
    if the transaction runs into a retry loop.
    """


    def update(self, data, raiseErrors=False, **kwargs):
        """Update an object with given data without using the built in
        transaction handling. This method calls setup which makes sure that
        we do not mark the object as _m_changed.
        """
        # prevent changes if tansaction is doomed
        t  = transaction.get()
        if t.isDoomed():
            return False

        # updated object data
        orgData = self.setup(data, dumpOriginalData=True)

        # update the data in MongoDB
        collection = self.__parent__.collection
        try:
            if self._version == 0:
                dump = self.__parent__.doInsertDump(self)
                self.__parent__.doInsert(collection, dump)
            else:
                dump = self.__parent__.doUpdateDump(self)
                self.__parent__.doUpdate(collection, dump, upsert=False)
        except (TypeError, pymongo.errors.ConnectionFailure,
            pymongo.errors.AutoReconnect, pymongo.errors.InvalidName) as e:
            # revert data, the transaction will try to store them later
            self.setup(orgData)
            if raiseErrors:
                raise e
            else:
                return False

        # mark _m_changed with False and not None like we do in load method
        # otherwise the item get processed again with our transaction handling
        self._m_changed = False
        return True


class MongoItemBase(MongoParentAware, SetupConvertDump):
    """Mongo item base class.

    Note: only MongoFieldProperty attributes get observed and updated by our
    transaction handling. If you change other attributes, you have to set
    _m_changed to True by yourself.

    Note: We do not support manipulators known from pymongo. We use a converter
    pattern for ding the convertion, see converters
    """

    # built in attrs
    _id = None
    _type = None
    _version = None
    _m_independent = False

    # built in skip and dump names
    _skipNames = []
    _dumpNames = ['_id', '_pid', '_type', '_version', '__name__',
                  'created', 'modified']

    # attrs with default values
    created = MongoFieldProperty(interfaces.IMongoItem['created'])
    modified = MongoFieldProperty(interfaces.IMongoItem['modified'])

    def __init__(self, data):
        """Initialize a mongo item with given data.

        This method will also make sure that the __parent__ and other values
        get set before we call setup.

        Note: if you enhance your own MongoItem __init__ method, don't forget
        to set_m_changed to None. Otherwise it will end in write access on
        each read.
        """
        # set given or None _pid (__parent__._id) reference
        self._pid = data.pop('_pid', None)
        # set given or None __parent__
        self.__parent__ = data.pop('__parent__', None)

        # set given or new _id
        _id = data.pop('_id', _marker)
        if _id is _marker:
            _id = bson.objectid.ObjectId()
        self.__dict__['_id'] = _id

        # set given or new _type if not already set as built-in
        if self._type is None:
            _type = data.pop('_type', self.__class__.__name__)
            if _type != self.__class__.__name__:
                raise TypeError("Wrong mongo item _type used")
            self.__dict__['_type'] = str(_type)

        # set given or 0 (zero) _version
        self.__dict__['_version'] = data.pop('_version', 0)

        # set given or new created datetime
        created = data.pop('created', _marker)
        if created is _marker:
            created = m01.mongo.now(UTC)
        self.__dict__['created'] = created

        # setup object with given (non IMongoItem) key/value data
        self.setup(data)

        # it is very important to set _m_changed to None, otherwise each read
        # access will end in a write access.
        self._m_changed = None

    def notifyRemove(self):
        """Notifies an item before dumped and removed from MongoDB."""
        pass


class SecureMongoItemBase(MongoItemBase):
    """Secure mongo item base class."""

    # built in skip and dump names
    _skipNames = []
    _dumpNames = ['_id', '_pid', '_type', '_version', '__name__',
                  'created', 'modified',
                  '_ppmrow', '_ppmcol',
                  '_prmrow', '_prmcol',
                  '_rpmrow', '_rpmcol',
                  ]

    def __init__(self, data):
        self._ppmrow = {}
        self._ppmcol = {}
        self._prmrow = {}
        self._prmcol = {}
        self._rpmrow = {}
        self._rpmcol = {}
        super(SecureMongoItemBase, self).__init__(data)


class MongoSubObjectBase(MongoParentAware, SetupConvertDump):
    """Mongo item base class.

    Note: only MongoFieldProperty attributes get observed and updated by our
    transaction handling. If you change other attributes, you have to set
    _m_changed to True by yourself.

    Note: We do not support manipulators known from pymongo. We use a converter
    pattern for ding the convertion, see converters
    """

    # built in attrs
    _type = None
    __m_changed = None

    # built in skip and dump names
    _skipNames = []
    _dumpNames = ['_type', '__name__']

    def __init__(self, data):
        """Initialize a mongo sub object with given data.

        This method will also make sure that the __name__ and __parent__ and
        other values get set before we call setup.

        Note: if you enhance your own MongoSubObjectBase __init__ method, don't
        forget to set_m_changed to None if you change value. Otherwise it will
        end in write access on each read access.
        """
        # set given or None __name__ and __parent__
        self.__name__ = data.pop('__name__', None)
        self.__parent__ = data.pop('__parent__', None)
        # set given or new _type if not already set as built-in
        if self._type is None:
            _type = data.pop('_type', self.__class__.__name__)
            if _type != self.__class__.__name__:
                raise TypeError("Wrong mongo item _type used")
            self.__dict__['_type'] = str(_type)
        # setup object with given (non IMongoItem) key/value data
        self.setup(data)
        # it is very important to set _m_changed to None, otherwise each read
        # access will end in a write access.
        self._m_changed = None

    def _m_changed():
        def get(self):
            return self.__m_changed
        def set(self, value):
            self.__m_changed = value
            if self.__parent__ is not None and value == True:
                self.__parent__._m_changed = True
        return property(get, set)

    _m_changed = _m_changed()


@implementer(interfaces.IMongoTransactionAware)
class MongoMappingBase(MongoItemAware, zope.location.location.Location):
    """Mongo mapping base class with thread save transaction and caching
    support.

    Note: right now we do not support manipulators. If we add suport for
    manipulators, check/change collection update and insert calls.

    Note: the databaseName and collectionName are not exposed in the API this
    is only use in our default implementation. You just have to make sure to
    define a collection and a cacheKey.

    Note: if you define a custom load method, you can use the built-in _type
    value where each Montoitem will use. By default the MongoItem uses it's
    class.__name__ as _type.

    """


    @property
    def collection(self):
        """Returns a mongodb collection"""
        raise NotImplementedError(
            "Subclass must implement collection attribute")

    @property
    def cacheKey(self):
        """Provide a thread local and unique cache key.

        Note: It's very important that you define a unique cache key which does
        not conflict with other instances or you will get into trouble because
        2 object whould begin the transaction before vote them. Then the first
        transaction will move the items to the transaction cache and the second
        object whould remove the transaction cache because of the same cache
        key before any vore call get processed which whould add items to the
        mongodb. The id(self) will make sure that we use unique cache keys per
        instance.
        """
        return '%s.%i.%i' % (self.collection.full_name, id(self),
            _thread.get_ident())

    def load(self, data):
        """Load data into a IMongoItem

        Since the default MongoContainerItem and MongoStorageItem are useless,
        we do not provide a default implementation.

        Implement this method if you need another mongo item. If you override
        the doLoad method which normaly calls this method, take care that your
        items get marked as not _m_changed. Otherwise it will end in write
        access on each read operation. Also see what else we do in doLoad
        method an do the same if needed. e.g. locate items etc.
        """
        raise NotImplementedError("Subclass must implement load")

    def doLoad(self, data):
        """Prepare data and load them into IMongoItem.

        This method calls the load method and will make sure that we
        inject a __parent__ and mark the loaded object as not _m_changed

        BIG NOTE: never ever load partial data or the transaction will write
        only this partial data back to mongo. This means never load data if
        they get loaded with the doFind method using a fields argument within
        a subset of the items fields!
        """
        # first validate data
        assert data['_id'] is not None
        # the __setitem__ or add methods are responsible that a _pid get
        # added. This means we can just assert them if an _mpid is given
        if self._mpid is not None:
            assert data['_pid'] == self._mpid

        __name__ = data.get('__name__')
        # make sure we never load removed data
        if __name__ in self._cache_removed:
            raise KeyError(__name__)
        # ensure that we never load objects by data twice, allways return our
        # objects from cache first, otherwise we will have two instances
        # representing the same object which will make troubles if we commit
        # our transaction
        obj = self._cache_loaded.get(__name__)
        # only load if not cached
        if obj is not None:
            # this only fails if we do not provide unique cacheKey
            assert obj.__parent__ is self
            assert obj._id == data['_id']
            # return the cached obj
            return obj
        # locate without to set _pid
        data['__parent__'] = self
        obj = self.load(data)
        # cache
        self._cache_loaded[obj.__name__] = obj
        # mark as NOT changed
        obj._m_changed = False
        # return located obj
        return obj

    # filter
    def doFilter(self, spec):
        """Base filter for all do* methods if not overriden"""
        if self._mpid is not None:
            spec['_pid'] = self._mpid
        return spec

    def doCountFilter(self, spec):
        """Filter for doCount method"""
        return self.doFilter(spec)

    def doFindOneFilter(self, spec):
        return self.doFilter(spec)

    def doFindFilter(self, spec):
        return self.doFilter(spec)

    def doBatchDataFilter(self, spec):
        return self.doFilter(spec)

    # dump methods (including validation)
    def doInsertDump(self, obj):
        assert obj._version == 0
        obj._version += 1
        data = obj.dump()
        assert data['_id'] is not None
        assert data['_version'] == 1
        if self._mpid is not None:
            assert data['_pid'] == self._mpid
        return data

    def doUpdateDump(self, obj):
        assert obj._version  > 0
        curVersion = obj._version
        obj._version += 1
        data = obj.dump()
        assert data['_id'] is not None
        if self._mpid is not None:
            assert data['_pid'] == self._mpid
        assert data['_version'] > 1
        # has version and not independent, check if not changed
        if not obj._m_independent:
            spec = {'_id': data['_id'], '_version': curVersion}
            if self.collection.find_one(spec) is None:
                raise ValueError(
                    "MongoItem version does not compare")
        return data

    def doRemoveDump(self, obj):
        assert obj._version  > 0
        curVersion = obj._version
        obj._version += 1
        data = obj.dump()
        assert data['_id'] is not None
        if self._mpid is not None:
            assert data['_pid'] == self._mpid
        assert data['_version'] > 1
        # has version and not independent, check if not changed
        if not obj._m_independent:
            spec = {'_id': data['_id'], '_version': curVersion}
            if self.collection.find_one(spec) is None:
                raise ValueError(
                    "MongoItem version does not compare")
        return data

    # mongo commiter
    def doCount(self, collection, spec=None, skipFilter=False):
        if spec is None:
            spec = bson.son.SON()
        if not skipFilter:
            spec = self.doCountFilter(spec)
        return collection.count_documents(spec)

    def doFindOne(self, collection, spec=None, fields=None, skipFilter=False):
        if spec is None:
            spec = bson.son.SON()
        if not skipFilter:
            spec = self.doFindOneFilter(spec)
        return collection.find_one(spec, fields)

    def doFind(self, collection, spec=None, fields=None, skipFilter=False):
        if spec is None:
            spec = bson.son.SON()
        if not skipFilter:
            spec = self.doFindFilter(spec)
        return collection.find(spec, fields)

    def doInsert(self, collection, data):
        if not isinstance(data, list):
            data = [data]
        for d in data:
            collection.insert_one(d)

    def doUpdate(self, collection, data, upsert=False):
        if not isinstance(data, list):
            data = [data]
        for d in data:
            # doUpdateDump is allowed to return None for block item update
            if d is not None:
                # very critical error, never update items without an _id
                assert d['_id'] is not None
                filter = {'_id': d['_id']}
                # skip _id with $set, OperationFailure: Mod on _id not allowed
                d.pop('_id', None)
                collection.update_one(filter, {'$set': d}, upsert=upsert)

    def doRemove(self, collection, data):
        if not isinstance(data, list):
            data = [data]
        for d in data:
            # doRemoveDump is allowed to return None for block item remove
            if d is not None:
                # very critical error, never remove items without an _id
                assert d['_id'] is not None
                collection.delete_one({'_id': d['_id']})

    # caches
    @property
    def _cache(self):
        """Thread local cache for objects."""
        return LOCAL.__dict__.setdefault(self.cacheKey, {})

    @property
    def _cache_added(self):
        return self._cache.setdefault('added', {})

    @property
    def _cache_loaded(self):
        return self._cache.setdefault('loaded', {})

    @property
    def _cache_removed(self):
        return self._cache.setdefault('removed', {})

    @property
    def _t_cache_insert(self):
        return self._cache.setdefault('insert', [])

    @property
    def _t_cache_update(self):
        return self._cache.setdefault('update', [])

    @property
    def _t_cache_remove(self):
        return self._cache.setdefault('remove', [])

    def resetLoadCache(self):
        """Reset thread local load caches."""
        _cache = self._cache
        if 'added' in _cache:
            del _cache['added']
        if 'loaded' in _cache:
            del _cache['loaded']
        if 'removed' in _cache:
            del _cache['removed']

    def resetTransactionCache(self):
        """Reset thread local transaction caches."""
        _cache = self._cache
        if 'insert' in _cache:
            del _cache['insert']
        if 'update' in _cache:
            del _cache['update']
        if 'remove' in _cache:
            del _cache['remove']

    def resetCache(self):
        """Reset thread local cache."""
        if self.cacheKey in LOCAL.__dict__:
            del LOCAL.__dict__[self.cacheKey]

    # IContainer API
    def __len__(self):
        # also take as removed and added cached items into account
        counts = self.doCount(self.collection)
        return counts - len(self._cache_removed) + len(self._cache_added)

    def items(self):
        # join transaction handling
        self.ensureTransaction()
        for data in self.doFind(self.collection):
            __name__ = data['__name__']
            if __name__ in self._cache_removed:
                # skip removed items
                continue
            obj = self._cache_loaded.get(__name__)
            if obj is None:
                try:
                    # load, locate and cache if not cached
                    obj = self.doLoad(data)
                except (KeyError, TypeError):
                    continue
            yield __name__, obj
        # also return keys/items not stored in MongoDB yet
        for k, v in list(self._cache_added.items()):
            yield k, v

    def keys(self):
        for data in self.doFind(self.collection, {}, ['__name__']):
            __name__ = data['__name__']
            if __name__ in self._cache_removed:
                # skip removed items
                continue
            else:
                yield __name__
        # also return keys not stored in MongoDB yet
        for k, v in list(self._cache_added.items()):
            yield k

    def values(self):
        # join transaction handling
        self.ensureTransaction()
        for data in self.doFind(self.collection):
            __name__ = data['__name__']
            if __name__ in self._cache_removed:
                # skip removed items
                continue
            obj = self._cache_loaded.get(__name__)
            if obj is None:
                try:
                    # load, locate and cache if not cached
                    obj = self.doLoad(data)
                except (KeyError, TypeError):
                    continue
            yield obj
        # also return items not stored in MongoDB yet
        for k, v in list(self._cache_added.items()):
            yield v

    def __iter__(self):
        """Return an iterator for the keys of the mapping object.
        """
        return iter(list(self.keys()))

    def __getitem__(self, key):
        """get item"""
        raise NotImplementedError("Subclass must implement __getitem__")

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError as e:
            pass
        return default

    def __delitem__(self, key):
        """delete item"""
        # load object
        obj = self[key]
        # notify remove
        obj.notifyRemove()
        # cleanup cache
        if key in self._cache_added:
            # remove new items
            del self._cache_added[key]
        else:
            # import pdb; pdb.set_trace()
            # or add to remove cache but only if not yet added
            self._cache_removed[key] = obj
        # also remove from loaded items
        if key in self._cache_loaded:
            del self._cache_loaded[key]

    def getBatchData(self, query=None, page=1, size=25, sortName=None,
        sortOrder=None, searchText=None, fields=None, skipFilter=False):
        """Returns batched mongo data, current page, total items and page size.

        Note: this method will not return loaded IMongoItems. If you need to
        load the given mongo data you can simply iterate the given cursor and
        call doLoad(data). The doLoad method also takes care of all the cached
        items. But take care, since mongo returns removed items and doesn't
        know about new added items till a transaction commits, this method
        could return bad items. And if you like to skip removed items from
        the curser by yourself, the page and pages counter could get messed up.
        If you like to bypass this troubles, I highly recommend only call
        doLoad based on cursor data if you didn't add or remove items from a
        container in the same transaction.
        (response.redirect or transaction.commit are your friends)

        Note: the page can get recalculated if the page position doesn't fit
        into the batch. This means the page in the reult can be different as
        the given page input. This is usefull if we use a larger page number
        as we realy have, e.g. after remove an item.

        Note: this method will call doBatchDataFilter which could enhance
        the given query specification

        Note: take care, this method will ignore cached (removed, added) items
        till we commit the transaction. You can eighter commit the transaction
        or use a response redirect in your request before using this method
        after remove or add items.

        NOTE: searchText is not used in this method. But this method signature
        allows to implement a search pattern in your own implementation.
        IF you use a pure mongo $text index, simply provide the right $text
        query part in the query argument. If you use elasticsearch or other
        text indexing server simply implement your own concept. This method
        signature makes it possible to implement generic search APIs in UI
        like we use for livesearch etc.
        """
        # start with server, this allows us to approve the page position
        if query is None:
            query = bson.son.SON()
        if not skipFilter:
            # skip batch data filter
            query = self.doBatchDataFilter(query)
        cursor = self.collection.find(query, fields)

        # get overall total based on query
        total = self.collection.count_documents({})
        # calculate pages
        pages = old_div(total,size)
        if pages == 0 or total % size:
            pages += 1
        # as next we approve our page position
        if page > pages:
            # restart with pages number as page which is the last page
            page = pages
            return self.getBatchData(query, page, size, sortName, sortOrder,
                searchText, fields, skipFilter)

        # calculate start size
        start = (page-1) * size
        cursor = cursor.skip(start)
        # sort result
        if sortName is not None:
            if sortOrder is None:
                if isinstance(sortName, six.string_types):
                    sortName = [(sortName, pymongo.ASCENDING)]
                elif not isinstance(sortName, list):
                    raise TypeError("if no sortOrder is specified, "
                                    "sortName must be an instance of list")
            cursor = cursor.sort(sortName, sortOrder)
        # limit result
        cursor = cursor.limit(size)
        # return data including probably adjusted page number
        return (cursor, page, pages, total)

    # transaction API
    def ensureTransaction(self):
        """Ensure that a transaction data manager observes our modification.

        Note: tis method is not exposed in our interface API.

        Note: we do not know if we return objects if they get modified. This
        means we observe all container which return items.

        Our transaction is not responsible for cleanup our thread local caches.
        We use an EndRequestEvent handler for cleanup our thread local cached
        items. Take care if you do testing, normaly the cache cleanup subscriber
        is not a part of the testing setup. But you can cleanup the caches by
        just call m01.mongo.clearThreadLocalCache(None) if you need to.
        """
        ensureMongoTransaction(self)

    def begin(self):
        """Prepare commit transaction handling.

        We move all relevant items from the load caches to the commit caches.
        This will give us empty load caches if the transactions runs into the
        retry loop.

        Note, prevent to load other mongo items which will run into ensure other
        transactions. This will fail if we join such new transactions during
        our commit phase.
        """
        self.resetTransactionCache()
        ci = self._t_cache_insert
        cu = self._t_cache_update
        dt = m01.mongo.now(UTC)
        for obj in list(self._cache_loaded.values()):
            if obj._m_changed:
                # set modified
                obj.modified = dt
                if obj._version == 0:
                    # has no version, add to insert cache
                    ci.append(self.doInsertDump(obj))
                else:
                    # has version, add to update cache
                    cu.append(self.doUpdateDump(obj))

        # add removed objects specification to transaction remove cache
        cr = self._t_cache_remove
        for obj in list(self._cache_removed.values()):
            cr.append(self.doRemoveDump(obj))

        # reset load cache which will support a clean retry
        self.resetLoadCache()

    def commit(self):
        """Start commit transaction handling."""
        pass

    def vote(self):
        """Process MongoDB insert and update calls.

        Note, if something fails, this could leave inconsistent data in MongoDB.

        """
        collection = self.collection
        # first remove removed items
        if len(self._t_cache_remove):
            self.doRemove(collection, self._t_cache_remove)
        # insert new items
        if len(self._t_cache_insert):
            self.doInsert(collection, self._t_cache_insert)
        # update existing items
        if len(self._t_cache_update):
            self.doUpdate(collection, self._t_cache_update, upsert=False)
        # reset transaction cache
        self.resetTransactionCache()

    def abort(self):
        """Abort transaction handling."""
        self.resetTransactionCache()
        self.resetLoadCache()
        self.resetCache()

    def finish(self):
        """Finish transaction commit."""
        self.resetTransactionCache()
        self.resetLoadCache()
        self.resetCache()


class MongoStorageBase(MongoMappingBase):
    """Mongo storage base class using mongo_id as __name__"""

    def __contains__(self, key):
        """Checks for an item by the given key without to load them"""
        if key in self._cache_removed:
            # NOTE: your implementation must make sure what should be done
            # for removed items, we just return False here
            return False
        elif self._cache_loaded.get(key):
            return True
        else:
            try:
                if PY2 and isinstance(key, newstr):
                    key_str = "{}".format(key)  # ObjectId requires a simple str
                else:
                    key_str = str(key)  # Standard conversion in Python 3

                data = self.doFindOne(self.collection,
                    {'_id': bson.objectid.ObjectId(key_str)})
                if data is not None:
                    return True
            except(pymongo.errors.InvalidId, TypeError) as e:
                pass

        return False

    def add(self, obj):
        """Add an item."""
        assert obj.__name__ is not None
        if obj.__name__ in self:
            # duplicated item
            raise KeyError(obj.__name__)
        # join transaction handling
        self.ensureTransaction()

        # set the parent
        obj = removeSecurityProxy(obj)
        # locate and set _pid if None
        obj.__parent__ = self

        # mark as changed
        obj._m_changed = True

        self._cache_added[obj.__name__] = obj
        self._cache_loaded[obj.__name__] = obj
        return obj.__name__

    def __getitem__(self, key):
        """get item"""
        if key in self._cache_removed:
            raise KeyError(key)
        obj = self._cache_loaded.get(key)
        if obj is None:
            try:
                if PY2 and isinstance(key, newstr):
                    key_str = "{}".format(key)  # ObjectId requires a simple str
                else:
                    key_str = str(key)  # Standard conversion in Python 3

                data = self.doFindOne(self.collection,
                    {'_id': bson.objectid.ObjectId(key_str)})
                if data is None:
                    raise KeyError(key)
            except(pymongo.errors.InvalidId, TypeError) as e:
                raise KeyError(key)
            obj = self.doLoad(data)
            # join transaction handling
            self.ensureTransaction()
        return obj


class MongoContainerBase(MongoMappingBase):
    """Mongo container base class using _id and __name__ as unique index."""

    def __contains__(self, key):
        """Checks for an item by the given key without to load them"""
        if key in self._cache_removed:
            # NOTE: your implementation must make sure what should be done
            # for removed items, we just return False here
            return False
        elif self._cache_loaded.get(key):
            return True
        else:
            try:
                data = self.doFindOne(self.collection, {'__name__': key})
                if data is not None:
                    return True
            except(pymongo.errors.InvalidId, TypeError) as e:
                pass

        return False

    def __setitem__(self, key, obj):
        """Set item"""
        if obj.__name__ is not None and obj.__name__ != key:
            raise KeyError("Key %r does not compare with object.__name__" % key,
                obj.__name__)
        if key in self:
            # duplicated item
            raise KeyError(key)
        # join transaction handling
        self.ensureTransaction()

        # set the name and parent
        obj = removeSecurityProxy(obj)
        # locate and set _pid if None
        obj.__parent__ = self
        obj.__name__ = key

        # mark as changed
        obj._m_changed = True

        self._cache_added[obj.__name__] = obj
        self._cache_loaded[obj.__name__] = obj

    def __getitem__(self, key):
        """get item"""
        if key in self._cache_removed:
            raise KeyError(key)
        obj = self._cache_loaded.get(key)
        if obj is None:
            try:
                data = self.doFindOne(self.collection, {'__name__': key})
                if data is None:
                    raise KeyError(key)
            except(pymongo.errors.InvalidId, TypeError) as e:
                raise KeyError(key)
            obj = self.doLoad(data)
            # join transaction handling
            self.ensureTransaction()
        return obj



###############################################################################
#
# Copyright (c) 2011 Projekt01 GmbH and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
###############################################################################
"""
$Id:$
"""
from __future__ import absolute_import
from past.builtins import cmp
from zope.interface import implementer
__docformat__ = "reStructuredText"

import zope.location
import zope.interface

from m01.mongo import interfaces


@implementer(interfaces.IMongoItemsData)
class MongoItemsData(zope.location.location.Location):
    """Mongo item list with _m_changed and location handling support."""


    def __init__(self, data=[]):
        self.clear()
        for obj in data:
            __name__, obj = self.locateAndReturn(obj)
            self.data[__name__] = obj
            self._order.append(__name__)

    def clear(self):
        """Clear data"""
        self.data = {}
        self._order = []

    def _m_changed():
        def get(self):
            raise AttributeError("_m_changed is a write only property")
        def set(self, value):
            if value == True:
                self.__parent__._m_changed = True
            else:
                raise ValueError("Can only dispatch True to __parent__")
        return property(get, set)

    _m_changed = _m_changed()
    def dump(self):
        items = []
        for obj in list(self.values()):
            if hasattr(obj, 'dump'):
                # dump mongo sub items
                items.append(obj.dump())
            else:
                # store simple value
                items.append(obj)
        return items

    def locate(self, value):
        """Locate an object with parent but only if locatable."""
        try:
            value.__parent__ = self
        except AttributeError as e:
            pass

    def locateAndReturn(self, obj):
        """Locate and return an object with parent but only if locatable."""
        self.locate(obj)
        return obj.__name__, obj

    def reLocate(self, __parent__):
        """Update location, this is needed if we use LocationProxy."""
        for item in list(self.values()):
            item.__parent__ = __parent__

    def order(self, orderNames):
        """Order items by the given name order"""
        if set(self._order) != set(orderNames):
            raise ValueError("Not all names used for order items")
        self._order = orderNames
        self._m_changed = True

    def fetch(self, idx):
        key = self._order[idx]
        return self.data[key]

    def append(self, item):
        """Append new item"""
        if item.__name__ in self._order:
            raise ValueError("Given item already appended", item.__name__)
        self._order.append(item.__name__)
        self.data[item.__name__] = item
        self.locate(item)
        self._m_changed = True

    def insert(self, i, item):
        """Insert new item at given order"""
        if item.__name__ in self._order:
            raise ValueError("Given item already inserted", item.__name__)
        self._order.insert(i, item.__name__)
        self.data[item.__name__] = item
        self.locate(item)
        self._m_changed = True

    def remove(self, item):
        """Remove an item"""
        del self.data[item.__name__]
        self._order.remove(item.__name__)
        self._m_changed = True

    def __len__(self):
        return len(self.data)

    def __contains__(self, key):
        return key in self.data

    def __getitem__(self, key):
        return self.data[key]

    def __setitem__(self, key, item):
        if key != item.__name__:
            raise KeyError("Only item __name__ are allowed as names")
        self.append(item)

    def __delitem__(self, key):
        item = self.data[key]
        self.remove(item)

    def get(self, key, default=None):
        """Get item by key"""
        return self.data.get(key, default)

    def keys(self):
        return self._order

    def values(self):
        for key in self._order:
            yield self.data[key]

    def items(self):
        for key in self._order:
            yield key, self.data[key]

    def __iter__(self):
        for key in self._order:
            yield self.data[key]

    def __cast(self, other):
        """Allows to compare other MongoItemsData lists."""
        if isinstance(other, MongoItemsData):
            return other.data
        else:
            return other

    def __cmp__(self, other):
        return cmp(self.data, self.__cast(other))

    def __repr__(self):
        """Represent this item as list of values"""
        return repr(list(self.data.values()))


@implementer(interfaces.IMongoListData)
class MongoListData(zope.location.location.Location):
    """Mongo item list with _m_changed support."""


    def __init__(self, data=[]):
        """Initialize given values in a new list

        IMPORTANT, it's absolutly required that we copy the values from the
        data list because it could be a global default value!. If we do not
        copy the data list, we would reference the default value from the
        schema field. eeek
        """
        self.data = list(data)

    def clear(self):
        """Clear data"""
        self.data = []

    def _m_changed():
        def get(self):
            raise AttributeError("_m_changed is a write only property")
        def set(self, value):
            if value == True:
                self.__parent__._m_changed = True
            else:
                raise ValueError("Can only dispatch True to __parent__")
        return property(get, set)

    _m_changed = _m_changed()
    def dump(self):
        items = []
        for obj in self:
            if hasattr(obj, 'dump'):
                # dump mongo sub items
                items.append(obj.dump())
            else:
                # store simple value
                items.append(obj)
        return items

    def append(self, item):
        self.data.append(item)
        self._m_changed = True

    def insert(self, i, item):
        self.data.insert(i, item)
        self._m_changed = True

    def pop(self, i=-1):
        rtn = self.data.pop(i)
        self._m_changed = True
        return rtn

    def remove(self, item):
        self.data.remove(item)
        self._m_changed = True

    def sort(self, *args, **kwargs):
        self.data.sort(*args, **kwds)
        self._m_changed = True

    def __len__(self):
        return len(self.data)

    def __contains__(self, item):
        return item in self.data

    def __getitem__(self, i):
        return self.data[i]

    def __setitem__(self, i, item):
        self.data[i] = item
        self._m_changed = True

    def __delitem__(self, i):
        del self.data[i]
        self._m_changed = True

    def __cast(self, other):
        """Allows to compare other MongoListData lists."""
        if isinstance(other, MongoListData):
            return other.data
        else:
            return other

    def __cmp__(self, other):
        return cmp(self.data, self.__cast(other))

    def __repr__(self):
        return repr(self.data)



###############################################################################
#
# Copyright (c) 2011 Projekt01 GmbH and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
###############################################################################
"""
$Id: performance.py 5178 2025-03-18 15:08:21Z felipe.souza $
"""
from __future__ import absolute_import
from __future__ import print_function
from builtins import range
from six.moves import range
__docformat__ = "reStructuredText"

import time
import transaction
from ZODB.DB import DB
from ZODB.DemoStorage import DemoStorage

import m01.stub.testing
import m01.mongo.testing


timeResult = None


def timeTest(function, counter=10000, *args, **kw):
    """timer"""
    res = []
    append = res.append
    def wrapper(*args, **kw):
        start_time = time.time()
        for i in range(counter):
            append(function(*args, **kw))
        total_time = time.time() - start_time
        global timeResult
        timeResult = total_time
        # commit transaction forces indexer collector indexing our objects
        transaction.commit()
        return res
    return wrapper(*args, **kw)

###############################################################################
#
# test methods
#
###############################################################################

def runSave(collection, amountOfObjects=1000):
    for i in range(amountOfObjects):
        data = {'title': u'Title %s' % i,
                'description': u'Description %s' % i}
        collection.save(data)


def runSetup(storage, amountOfObjects=1000):
    """Only setup and dump object without any DB operation."""
    data = {'title': u'Title',
            'description': u'Description'}
    for i in range(amountOfObjects):
        m01.mongo.testing.SampleStorageItem(data)


def runSetupAndDump(storage, amountOfObjects=1000):
    """Only setup and dump object without any DB operation."""
    data = {'title': u'Title',
            'description': u'Description'}
    for i in range(amountOfObjects):
        item = m01.mongo.testing.SampleStorageItem(data)
        item.dump()


def runAdd(storage, amountOfObjects=1000):
    data = {'title': u'Title',
            'description': u'Description'}
    for i in range(amountOfObjects):
        item = m01.mongo.testing.SampleStorageItem(data)
        storage.add(item)


def runAddWithCommit(storage, amountOfObjects=1000):
    data = {'title': u'Title',
            'description': u'Description'}
    item = m01.mongo.testing.SampleStorageItem(data)
    for i in range(amountOfObjects):
        item = m01.mongo.testing.SampleStorageItem(data)
        storage.add(item)
    transaction.commit()


def runAddWithSingleCommit(storage, amountOfObjects=1000):
    data = {'title': u'Title',
            'description': u'Description'}
    item = m01.mongo.testing.SampleStorageItem(data)
    for i in range(amountOfObjects):
        item = m01.mongo.testing.SampleStorageItem(data)
        storage.add(item)
        transaction.commit()


def runObjectModified(obj, amountOfObjects=1000):
    for i in range(amountOfObjects):
        obj.title = u'Title %s' % i
        transaction.commit()


def runUpdate(storage, obj, amountOfObjects=1000):
    for i in range(amountOfObjects):
        data = {'title': u'Title %s' % i}
        obj.update(data)


def mongoCollections(storage, amountOfObjects=1000):
    for i in range(amountOfObjects):
        data = {'title': u'Title %s' % i,
                'description': u'Description %s' % i}
        collection.save(data)


###############################################################################
#
# performance test
#
###############################################################################

def getMongo():
    client = m01.mongo.testing.getTestClient()
    # tear down old database probably left over from aborted tests
    client.drop_database(m01.mongo.testing.TEST_DB_NAME)
    # setup new database
    mdb = client[m01.mongo.testing.TEST_DB_NAME]
    collection = mdb['performance']
    return mdb, collection


def runTest(repeatTimes, amountOfObjects):
    """Test methods

    Note, the ZODB is here for historical reason and shows that it is
    possible to store a container in ZODB whcih contains objects stored in
    mongodb. The transaction can handle the mixed environment.

    """

    # first setup mongo test database

    print("")
    print("Run test with")
    print("-------------")
    print("- %i x repeat tests" % repeatTimes)
    print("- %i objects" % amountOfObjects)

    # setup ZODB and storage
    zodb = DB(DemoStorage())
    conn = zodb.open()
    storage = m01.mongo.testing.SampleStorage()
    conn.root()['storage'] = storage
    transaction.commit()
    conn.close()

    # start test
    # setup mongo db and collection
    mdb, collection = getMongo()
    conn = zodb.open()
    storage = conn.root()['storage']
    timeTest(runSave, repeatTimes, collection, amountOfObjects)
    conn.close()
    print("save:  %.2f s" % timeResult)

    # setup mongo
    mdb, collection = getMongo()
    conn = zodb.open()
    storage = conn.root()['storage']
    timeTest(runSetup, repeatTimes, storage, amountOfObjects)
    conn.close()
    print("setup: %.2f s" % timeResult)

    # setup mongo
    mdb, collection = getMongo()
    conn = zodb.open()
    storage = conn.root()['storage']
    timeTest(runSetupAndDump, repeatTimes, storage, amountOfObjects)
    conn.close()
    print("setup and dump: %.2f s" % timeResult)

    # setup mongo
    mdb, collection = getMongo()
    conn = zodb.open()
    storage = conn.root()['storage']
    timeTest(runAdd, repeatTimes, storage, amountOfObjects)
    transaction.abort()
    conn.close()
    print("add:  %.2f s" % timeResult)

    # setup mongo
    mdb, collection = getMongo()
    conn = zodb.open()
    storage = conn.root()['storage']
    timeTest(runAddWithCommit, repeatTimes, storage, amountOfObjects)
    conn.close()
    print("add with commit:  %.2f s" % timeResult)

    # setup mongo
    mdb, collection = getMongo()
    conn = zodb.open()
    storage = conn.root()['storage']
    timeTest(runAddWithSingleCommit, repeatTimes, storage, amountOfObjects)
    conn.close()
    print("add with single commit:  %.2f s" % timeResult)

    # time object modified event
    # setup mongo
    mdb, collection = getMongo()
    conn = zodb.open()
    storage = conn.root()['storage']
    data = {'title': u'Title modified',
            'description': u'Description modified'}
    item = m01.mongo.testing.SampleStorageItem(data)
    storage.add(item)
    transaction.commit()
    timeTest(runObjectModified, repeatTimes, item, amountOfObjects)
    conn.close()
    print("object modified with single commit: %.2f s" % timeResult)

    # time object modified event
    # setup mongo
    mdb, collection = getMongo()
    conn = zodb.open()
    storage = conn.root()['storage']
    data = {'title': u'Title modified',
            'description': u'Description modified'}
    item = m01.mongo.testing.SampleStorageItem(data)
    storage.add(item)
    transaction.commit()
    timeTest(runUpdate, repeatTimes, storage, item, amountOfObjects)
    conn.close()
    print("object update: %.2f s" % timeResult)


def createCollection(mdb, amount):
    """Used for check how many collection we can add."""
    for i in range(amount):
        mdb.create_collection('collection-%i' % i)
        col = mdb['collection-%i' % i]
        for ii in range(25):
            col.ensure_index('idx-%i' % ii)


def dropCollection(mdb, amount):
    """Used for check how many collection we can add."""
    for i in range(amount):
        mdb.drop_collection('collection-%i' % i)


def runCollectionTest(amount):
    # first setup mongo test database
    setUpMongo()

    print("")
    print("Create collections")
    print("------------------")
    print("- %i collections" % amount)
    print("")
    # setup mongo
    mdb, collection = getMongo()
    timeTest(createCollection, 1, mdb, amount)
    print("create collection:  %.2f s" % timeResult)
    timeTest(dropCollection, 1, mdb, amount)
    print("drop collection:  %.2f s" % timeResult)


def main():
    error = None
    try:
        # m01.stub.testing.startMongoServer()
        m01.mongo.testing.setUpStubMongo()
        runTest(1, 100)
        runTest(1, 10000)
        # check the amount of collections. Note there is a limit of 24000
        # namspaces per collection. Each collection and index are counted as a
        # namespace. If you need more, you need to set the --nssize paramter.
        # You also need to use 64bit machine for test the namspace limit
        #runCollectionTest(24000)
    except Exception as e:
        error = e
    else:
        # m01.stub.testing.stopMongoServer(5.0)
        m01.mongo.testing.tearDownStubMongo()
    if error is not None:
        raise error



###############################################################################
#
# Copyright (c) 2011 Projekt01 GmbH and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
###############################################################################
"""
$Id:$
"""
from __future__ import absolute_import
from builtins import map
from zope.interface import implementer
from builtins import object
from six.moves import map
__docformat__ = "reStructuredText"

import zope.interface
import zope.component
import zope.publisher.interfaces
from zope.publisher.interfaces.http import IHTTPRequest
from zope.publisher.interfaces import IDefaultViewName
from zope.publisher.interfaces import NotFound
from zope.publisher.interfaces.browser import IBrowserPublisher

from m01.mongo import interfaces


class BrowserDefaultMixin(object):
    """Browser default mixin class.

    Note: we don't use the @@ prefix for IDefaultViewName because @@ prefixed
    view names force to bypass our IBrowserPublisher traverser and use the view
    namespace for lookup views.
    """

    def browserDefault(self, request):
        """See zope.publisher.browser.interfaces.IBrowserPublisher"""
        view_name = zope.component.getSiteManager(None).adapters.lookup(
            list(map(zope.interface.providedBy, (self.context, request))),
                IDefaultViewName)
        if view_name is None:
            raise zope.component.ComponentLookupError(
                "Couldn't find default view name", self.context, request)
        return self.context, (view_name,)


@implementer(IBrowserPublisher)
class MongoTraverserMixin(BrowserDefaultMixin):
    """MongoTraverser mixin class"""


    def __init__(self, context, request):
        self.context = context
        self.request = request

    def publishTraverse(self, request, name):
        """See zope.publisher.interfaces.IPublishTraverse"""
        raise NotImplementedError("Subclass must implement publishTraverse")


class MongoStorageTraverser(MongoTraverserMixin):
    """MongoStorageTraverser"""

    zope.component.adapts(interfaces.IMongoStorage, IHTTPRequest)

    def publishTraverse(self, request, name):
        """See zope.publisher.interfaces.IPublishTraverse"""

        # check storage if we get a valid MongoDB object id.
        if len(name) == 12 or len(name) == 24:
            try:
                return self.context[name]
            except KeyError:
                pass

        view = zope.component.queryMultiAdapter((self.context, request),
            name=name)
        if view is not None:
            return view

        raise zope.publisher.interfaces.NotFound(self.context, name, request)


class MongoContainerTraverser(MongoTraverserMixin):
    """MongoContainerTraverser"""

    zope.component.adapts(interfaces.IMongoContainer, IHTTPRequest)

    def publishTraverse(self, request, name):
        """See zope.publisher.interfaces.IPublishTraverse"""

        # Note: we changed the order of lookup because we do not like to
        # force the storage to lookup MongoDB if we need a view. But take
        # care, this means a view name will override a item key if you don't
        # use a @@ view marker prefix.
        view = zope.component.queryMultiAdapter((self.context, request),
            name=name)
        if view is not None:
            return view

        # this will force a MongoDB lookup
        try:
            return self.context[name]
        except KeyError:
            pass

        raise zope.publisher.interfaces.NotFound(self.context, name, request)


@implementer(IBrowserPublisher)
class MongoItemsDataTraverser(object):
    """MongoItemsTraverser"""

    zope.component.adapts(interfaces.IMongoItemsData, IHTTPRequest)

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def publishTraverse(self, request, name):
        """See zope.publisher.interfaces.IPublishTraverse"""
        try:
            return self.context[name]
        except TypeError as e:
            pass

        raise zope.publisher.interfaces.NotFound(self.context, name, request)
